# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""SpidR model implementation."""

import math
from functools import partial

import torch
import torch.nn.functional as F
from torch import Tensor

from spidr.config import SpidRConfig
from spidr.models.dinosr import DinoSR
from spidr.models.metrics import perplexity


def exp_ema_scheduler(step: int, start_decay: float, timescale: float, threshold: float) -> float:
    decay = 1 - (1 - start_decay) * math.exp(-step / timescale)
    return decay if 1 - decay > threshold else 1


class SpidR(DinoSR):
    def __init__(self, cfg: SpidRConfig | None = None) -> None:
        if cfg is None:
            cfg = SpidRConfig()
        super().__init__(cfg)
        self.ema_scheduler = partial(
            exp_ema_scheduler,
            start_decay=cfg.ema_start_decay,
            timescale=cfg.ema_timescale,
            threshold=cfg.ema_threshold,
        )

    def get_codebooks(
        self,
        waveform: Tensor,
        *,
        attention_mask: Tensor | None = None,
        onehot: bool = False,
    ) -> list[Tensor | None]:
        x = self.feature_extractor(waveform)
        x = self.feature_projection(x)
        preds: list[Tensor | None] = [None] * (len(self.student.layers) - self.num_codebooks)
        for i, y in enumerate(self.student.get_intermediate_outputs(x, attention_mask)[-self.num_codebooks :]):
            pred = self.heads[i](y).float().exp().squeeze()
            if onehot:
                pred = F.one_hot(pred.argmax(dim=-1), pred.size(-1))
            preds.append(pred)
        return preds

    def forward(
        self, waveforms: Tensor, *, mask: Tensor | None = None, attention_mask: Tensor | None = None
    ) -> tuple[Tensor, dict[str, Tensor]]:
        feats = self.feature_extractor(waveforms)
        feats = self.feature_projection(feats)
        x = feats.clone()
        x = self.projection_dropout(x)
        if mask is not None:
            x = torch.where(mask.unsqueeze(-1), self.mask_embedding.to(x.dtype).expand_as(x), x)
        else:
            mask = torch.ones((x.shape[0], x.shape[1]), dtype=torch.bool, device=x.device)
        mask_indices = torch.nonzero(mask, as_tuple=True)
        log_preds = [
            self.heads[i](y[mask_indices])
            for i, y in enumerate(self.student.get_intermediate_outputs(x, attention_mask)[-self.num_codebooks :])
        ]

        with torch.no_grad():
            targets = self.teacher.get_intermediate_outputs(feats, attention_mask)[-self.num_codebooks :]
            targets = [F.instance_norm(tl.float().transpose(1, 2)).transpose(1, 2)[mask_indices] for tl in targets]

        losses = torch.zeros(log_preds[0].shape[0], device=x.device)
        target_ppl, pred_ppl = torch.zeros((), device=x.device), torch.zeros((), device=x.device)
        for i, (log_pred, target) in enumerate(zip(log_preds, targets, strict=True)):
            onehot_target = self.codebooks[i](target)
            target_ppl += perplexity(onehot_target)
            pred_ppl += perplexity(log_pred.exp())
            losses += torch.sum(-onehot_target * log_pred, dim=-1)

        return (losses / self.num_codebooks), {
            "target_ppl": target_ppl / self.num_codebooks,
            "pred_ppl": pred_ppl / self.num_codebooks,
        }



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Environment setup."""

import logging
import os
import random
import subprocess
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import torch
import torch._appdirs
import torch._inductor.config
from torch import distributed as dist

from spidr.tools import init_logger

logger = logging.getLogger()


@dataclass(frozen=True)
class DistributedEnvironment:
    global_rank: int
    local_rank: int
    world_size: int
    master_port: int
    master_addr: str


def distributed_environment() -> DistributedEnvironment:
    if os.getenv("LOCAL_RANK") is not None:  # torchrun job
        env = DistributedEnvironment(
            global_rank=int(os.environ["RANK"]),
            local_rank=int(os.environ["LOCAL_RANK"]),
            world_size=int(os.environ["WORLD_SIZE"]),
            master_port=int(os.environ["MASTER_PORT"]),
            master_addr=os.environ["MASTER_ADDR"],
        )
    elif os.getenv("SLURM_JOB_ID") is not None:  # Slurm job
        scontrol_cmd = ["scontrol", "show", "hostnames", os.environ["SLURM_JOB_NODELIST"]]
        master_addr = subprocess.check_output(scontrol_cmd, text=True).split()[0]
        master_port = random.Random(int(os.environ["SLURM_JOB_ID"])).randint(20_000, 60_000)
        env = DistributedEnvironment(
            global_rank=int(os.environ["SLURM_PROCID"]),
            local_rank=int(os.environ["SLURM_LOCALID"]),
            world_size=int(os.environ["SLURM_NTASKS"]),
            master_port=master_port,
            master_addr=master_addr,
        )
    else:
        env = DistributedEnvironment(  # Single GPU job
            global_rank=0,
            local_rank=0,
            world_size=1,
            master_port=random.Random(-1).randint(20_000, 60_000),
            master_addr="127.0.0.1",
        )
    os.environ["RANK"] = str(env.global_rank)
    os.environ["LOCAL_RANK"] = str(env.local_rank)
    os.environ["WORLD_SIZE"] = str(env.world_size)
    os.environ["MASTER_PORT"] = str(env.master_port)
    os.environ["MASTER_ADDR"] = env.master_addr
    return env


def setup_distributed() -> None:
    env = distributed_environment()
    logger.debug("World size %s, local rank %s, global rank %s", env.world_size, env.local_rank, env.global_rank)
    dist.init_process_group(backend="nccl")
    logger.debug("DDP initialized")


def setup_environment(**kwargs: str) -> None:
    if "SLURM_JOB_ID" in os.environ and (Path("/scratch") / os.environ["SLURM_JOB_ID"]).is_dir():
        cache = Path("/scratch") / os.environ["SLURM_JOB_ID"]
    else:
        cache = Path(torch._appdirs.user_cache_dir(appname="spidr"))
    cache.mkdir(exist_ok=True, parents=True)
    kwargs["TMPDIR"] = str(cache / "tmp")
    kwargs["TORCHINDUCTOR_CACHE_DIR"] = str(cache / "torchinductor")
    kwargs["TRITON_HOME"] = str(cache)

    for name, value in kwargs.items():
        if os.getenv(name) != str(value):
            os.environ[name] = str(value)
            logger.debug("Setting env: %s=%s", name, value)


def set_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def setup_pytorch(*, use_deterministic: bool) -> None:
    torch.set_float32_matmul_precision("highest")
    torch.backends.cuda.matmul.allow_tf32 = False
    torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = True
    torch.backends.cudnn.benchmark = False  # Needed because of dynamic input size
    torch.autograd.set_detect_anomaly(mode=False, check_nan=True)
    if use_deterministic:
        torch.use_deterministic_algorithms(mode=True)
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
        torch._inductor.config.fallback_random = True


def setup_training(seed: int, *, use_deterministic: bool = False) -> None:
    init_logger()
    setup_distributed()
    set_seed(seed + dist.get_rank())
    setup_pytorch(use_deterministic=use_deterministic)
    setup_environment()



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Checkpoint management for training runs."""

import logging
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import torch
from torch import GradScaler, nn
from torch.nn.modules.utils import consume_prefix_in_state_dict_if_present
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

from spidr.tools import MetricsQueue

logger = logging.getLogger()


def remove_param_group(optim: Optimizer, index: int) -> None:
    for param in optim.param_groups[index]["params"]:
        if param in optim.state:
            del optim.state[param]
    del optim.param_groups[index]


def find_checkpoints(folder: Path) -> list[Path]:
    paths = sorted(folder.glob("step_*.pt"), key=lambda x: int(x.stem.removeprefix("step_")))
    if (folder / "final.pt").is_file():
        paths.append(folder / "final.pt")
    return paths


@dataclass
class Checkpoint:
    model: nn.Module
    optimizer: Optimizer
    scheduler: LRScheduler
    scaler: GradScaler
    epoch: torch.Tensor
    step: torch.Tensor


class Checkpointer:
    def __init__(
        self,
        folder: Path,
        interval: int,
        keep_latest: int = -1,
        *,
        disable: bool = False,
        start: int = 0,
    ) -> None:
        self.folder = Path(folder)
        self.folder.mkdir(parents=True, exist_ok=True)
        self.interval = interval
        self.keep_latest = keep_latest
        self.disable = disable or interval <= 0
        self._path: Path | None = None
        self._state: Checkpoint
        self._metrics_queue = MetricsQueue(self.folder / "metrics.jsonl", interval, start=start)

    @property
    def step(self) -> torch.Tensor:
        return self._state.step

    @property
    def epoch(self) -> torch.Tensor:
        return self._state.epoch

    @property
    def last(self) -> Path | None:
        if self._path is not None:
            return self._path
        if len(checkpoints := find_checkpoints(self.folder)) > 0:
            self._path = checkpoints[-1]
            return self._path
        return None

    @property
    def metrics(self) -> Path:
        return self.folder / "metrics.jsonl"

    def init_state(self, model: nn.Module, optimizer: Optimizer, scheduler: LRScheduler, scaler: GradScaler) -> None:
        epoch, step = torch.zeros(1, dtype=torch.long), torch.zeros(1, dtype=torch.long)
        ckpt = Checkpoint(model=model, optimizer=optimizer, scheduler=scheduler, scaler=scaler, epoch=epoch, step=step)
        self._state = ckpt

    def _purge_old_checkpoints(self) -> None:
        if self.keep_latest <= 0:
            return
        if len(checkpoints := find_checkpoints(self.folder)) <= self.keep_latest:
            return
        to_delete = checkpoints[: -self.keep_latest]
        for path in to_delete:
            logger.info("Deleting old checkpoint %s", path)
            path.unlink()

    def _save_to_path(self, path: Path, step: int, epoch: int) -> bool:
        if self.disable:
            return False
        self._state.step.fill_(step)
        self._state.epoch.fill_(epoch)
        torch.save(
            {
                "model": self._state.model.state_dict(),
                "optimizer": self._state.optimizer.state_dict(),
                "scheduler": self._state.scheduler.state_dict(),
                "scaler": self._state.scaler.state_dict(),
                "epoch": self.epoch,
                "step": self.step,
            },
            path,
        )
        self._path = path
        self._purge_old_checkpoints()
        return True

    def save(self, step: int, epoch: int, *, force: bool = False) -> bool:
        should_save = (not self.disable) and (force or step % self.interval == 0)
        if not should_save:
            return False
        return self._save_to_path(self.folder / f"step_{step}.pt", step, epoch)

    def save_final(self, step: int, epoch: int) -> bool:
        return self._save_to_path(self.folder / "final.pt", step, epoch)

    def _load_from_path(self, path: Path) -> bool:
        if self.disable or not path.is_file():
            return False
        if path.name == "final.pt":
            logger.warning("Loading final checkpoint")
        ckpt = torch.load(path, map_location=torch.device("cpu"), weights_only=True)
        self._state.step = ckpt["step"]
        self._state.epoch = ckpt["epoch"]
        self._state.scheduler.load_state_dict(ckpt["scheduler"])
        self._state.scaler.load_state_dict(ckpt["scaler"])

        consume_prefix_in_state_dict_if_present(ckpt["model"], "module.")
        self._state.model.load_state_dict(ckpt["model"])
        if self._state.model.current_step >= self._state.model.freeze_step:
            self._state.model.freeze_extractor()
            if len(self._state.optimizer.param_groups) > 1:
                remove_param_group(self._state.optimizer, 1)
        if len(self._state.optimizer.param_groups) != len(ckpt["optimizer"]["param_groups"]):
            raise ValueError("Inconsistency across current and checkpoint's param_groups")
        self._state.optimizer.load_state_dict(ckpt["optimizer"])
        return True

    def load_existing_run(self) -> bool:
        if (path := self.last) is None:
            return False
        return self._load_from_path(path)

    def load(self, step: int) -> bool:
        return self._load_from_path(self.folder / f"step_{step}.pt")

    def find_new_metrics(self) -> Generator[dict[str, Any], None, None]:
        yield from self._metrics_queue()



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Optimization utilities."""

import math

import torch
from torch import GradScaler
from torch.optim import AdamW, Optimizer
from torch.optim.lr_scheduler import CosineAnnealingLR, LambdaLR, LinearLR, LRScheduler, SequentialLR

from spidr.config import OptimizerConfig


def build_scheduler(opt: Optimizer, cfg: OptimizerConfig) -> LRScheduler:
    init_lr_scale = cfg.init_lr_scale if cfg.warmup_steps > 0 else 1
    decay: LRScheduler
    if cfg.scheduler == "tristage":
        warmup = LinearLR(opt, start_factor=init_lr_scale, total_iters=cfg.warmup_steps)
        hold = LinearLR(opt, start_factor=1.0, total_iters=cfg.hold_steps)
        decay = LambdaLR(opt, lambda step: math.exp(math.log(cfg.final_lr_scale) * step / cfg.decay_steps))
        return SequentialLR(opt, [warmup, hold, decay], [cfg.warmup_steps, cfg.hold_steps + cfg.warmup_steps])
    if cfg.scheduler == "cosine":
        warmup = LinearLR(opt, start_factor=init_lr_scale, total_iters=cfg.warmup_steps)
        decay = CosineAnnealingLR(opt, cfg.max_steps - cfg.warmup_steps, cfg.final_lr_scale * cfg.lr)
        return SequentialLR(opt, [warmup, decay], [cfg.warmup_steps])
    if cfg.scheduler == "rsqrt":
        warmup = LinearLR(
            opt,
            start_factor=init_lr_scale,
            end_factor=1 / math.sqrt(1 + cfg.rsqrt_shift / cfg.rsqrt_timescale),
            total_iters=cfg.warmup_steps,
        )
        hold = LinearLR(opt, start_factor=1.0, total_iters=cfg.hold_steps)
        decay = LambdaLR(opt, lambda step: 1 / math.sqrt(1 + (step + cfg.rsqrt_shift) / cfg.rsqrt_timescale))
        return SequentialLR(opt, [warmup, hold, decay], [cfg.warmup_steps, cfg.hold_steps + cfg.warmup_steps])
    if cfg.scheduler == "constant":
        warmup = LinearLR(opt, start_factor=init_lr_scale, total_iters=cfg.warmup_steps)
        hold = LinearLR(opt, start_factor=1.0, total_iters=cfg.max_steps)
        return SequentialLR(opt, [warmup, hold], [cfg.warmup_steps])
    raise ValueError(f"Unknown scheduler: {cfg.scheduler}")


def build_optimizer(model: torch.nn.Module, cfg: OptimizerConfig) -> tuple[AdamW, GradScaler, LRScheduler]:
    if cfg.dtype not in {"float32", "float16", "bfloat16"}:
        raise ValueError(cfg.dtype)
    if cfg.dtype == "bfloat16" and torch.cuda.is_available() and torch.cuda.get_device_capability() < (8, 0):
        raise ValueError("Cannot use bfloat16 on this GPU (V100?), try again with float16")

    param_groups: list[dict[str, list[torch.nn.Parameter]]] = [{"params": []}, {"params": []}]
    for name, param in model.named_parameters():
        if any(name.startswith(exclude) for exclude in cfg.exclude_from_optimizer):
            continue
        group = 1 if any(name.startswith(freeze) for freeze in cfg.to_freeze) else 0
        param_groups[group]["params"].append(param)

    optimizer = AdamW(param_groups, lr=cfg.lr, weight_decay=cfg.weight_decay, betas=cfg.betas, eps=cfg.eps, fused=True)
    scaler = GradScaler("cuda", enabled=cfg.mixed_precision)
    scheduler = build_scheduler(optimizer, cfg)
    return optimizer, scaler, scheduler



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Write manifest files."""

import os
import tarfile
from pathlib import Path

import polars as pl
from tqdm import tqdm

from spidr.data.utils import bytes_from_archive, num_samples


class CompressedArchiveError(Exception):
    """Archive must be uncompressed to be read."""


class NoAudioFileError(Exception):
    """No audio file found in the archive."""

    def __init__(self, archive: Path | str, file_extension: str) -> None:
        super().__init__(f"No file found in {archive} with extension `{file_extension}`.")


def write_manifest(dataset: Path | str, output: Path | str, file_extension: str = ".wav") -> None:
    """Write a manifest file containing the file paths and their number of samples.

    Each line contains the absolute path of the file and its number of samples.
    """
    files_info = [
        {"path": str(name), "num_samples": num_samples(name)}
        for name in tqdm(list(Path(dataset).resolve().rglob(f"*{file_extension}")))
    ]
    (
        pl.from_dicts(files_info)
        .lazy()
        .with_columns(pl.col("path").str.strip_suffix(file_extension).str.split(os.sep).list.get(-1).alias("fileid"))
        .select("fileid", "path", "num_samples")
        .sink_csv(output)
    )


def write_manifest_fairseq(dataset: Path | str, output: Path | str, file_extension: str = ".wav") -> None:
    """Write a manifest file containing the file paths and their number of samples.

    First line is the root directory of the dataset.
    Each line contains the relative path of the file and its number of samples.
    """
    lines = [str(Path(dataset).resolve())]
    paths = list(Path(dataset).rglob(f"*{file_extension}"))
    lines += [f"{name.relative_to(dataset)}\t{num_samples(name)}" for name in tqdm(paths)]
    Path(output).write_text("\n".join(lines) + "\n", encoding="utf-8")


def write_manifest_tar(dataset: Path | str, output: Path | str, file_extension: str = ".wav") -> None:
    if not tarfile.is_tarfile(dataset):
        raise ValueError(dataset)
    with tarfile.open(dataset, mode="r") as tar_file:
        if tar_file.mode != "r":
            raise CompressedArchiveError
        infolist = tar_file.getmembers()
    bytes_info = [
        {
            "path": info.name,
            "num_samples": num_samples(bytes_from_archive(dataset, info.offset_data, info.size)),
            "byte_offset": info.offset_data,
            "byte_size": info.size,
        }
        for info in tqdm(infolist)
        if not info.isdir() and info.name.endswith(file_extension)
    ]
    if not bytes_info:
        raise NoAudioFileError(dataset, file_extension)
    (
        pl.from_dicts(bytes_info)
        .lazy()
        .with_columns(
            pl.col("path").str.strip_suffix(file_extension).str.split(os.sep).list.get(-1).alias("fileid"),
            pl.lit(dataset).alias("archive"),
        )
        .select("fileid", "path", "num_samples", "archive", "byte_offset", "byte_size")
        .sink_csv(output)
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Write manifest files.")
    parser.add_argument("dataset", type=Path, help="Path to the dataset directory or uncompressed tar file.")
    parser.add_argument("output", type=Path, help="Path to the output manifest file.")
    parser.add_argument("--ext", type=str, default=".wav", help="Extension of audio files. (default: %(default)s)")
    parser.add_argument("--fairseq", action="store_true", help="Write a Fairseq-style TSV manifest, instead of CSV.")
    args = parser.parse_args()

    root = args.dataset.resolve()
    if root.is_dir():
        (write_manifest_fairseq if args.fairseq else write_manifest)(root, args.output, args.ext)
    elif root.suffix == ".tar":
        write_manifest_tar(root, args.ext)
    else:
        raise parser.error(f"{root} is not a directory or a tar file.")



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Model components for DinoSR and SpidR."""

import math

import torch
from torch import Tensor, nn
from torch import distributed as dist
from torch.nn import functional as F

from spidr.config import DinoSRConfig


class LayerNorm(nn.LayerNorm):
    """Layer norm with transpose."""

    def forward(self, x: Tensor) -> Tensor:
        x = x.transpose(-2, -1)
        x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        return x.transpose(-2, -1)


class ConvLayerBlock(nn.Module):
    """Convolution unit of FeatureExtractor."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        layer_norm: nn.Module | None,
        *,
        bias: bool,
    ) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.layer_norm = layer_norm
        self.conv = nn.Conv1d(in_channels, out_channels, kernel_size, stride, bias=bias)
        self.init_weights()

    def init_weights(self) -> None:
        nn.init.kaiming_normal_(self.conv.weight)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        if self.layer_norm is not None:
            x = self.layer_norm(x)
        return F.gelu(x)


class FeatureExtractor(nn.Module):
    """Extract features from audio."""

    def __init__(self, conv_layers: nn.ModuleList) -> None:
        super().__init__()
        self.conv_layers = conv_layers

    def forward(self, x: Tensor) -> Tensor:
        x = x.unsqueeze(1)  # (batch, channel==1, frame)
        for layer in self.conv_layers:
            x = layer(x)  # (batch, feature, frame)
        return x.transpose(1, 2)  # (batch, frame, feature)


class FeatureProjection(nn.Module):
    """Projects features to encoder dimension."""

    def __init__(self, in_features: int, out_features: int) -> None:
        super().__init__()
        self.layer_norm = nn.LayerNorm(in_features)
        self.projection = nn.Linear(in_features, out_features)

    def forward(self, x: Tensor) -> Tensor:
        x = self.layer_norm(x)
        return self.projection(x)


class ConvPositionalEmbedding(nn.Module):
    """Positional embedding which is placed at the beginning of Transformer."""

    def __init__(self, embed_dim: int, kernel_size: int, groups: int, depth: int) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.kernel_size = kernel_size
        self.groups = groups
        self.depth = depth
        self.layer_kernel_size = max(3, kernel_size // depth)
        self.num_remove = 1 if self.layer_kernel_size % 2 == 0 else 0
        self.layer_norm = LayerNorm(embed_dim, elementwise_affine=False)
        padding = self.layer_kernel_size // 2
        self.convs = nn.ModuleList(
            [
                nn.Conv1d(embed_dim, embed_dim, self.layer_kernel_size, padding=padding, groups=groups)
                for _ in range(depth)
            ]
        )

    def forward(self, x: Tensor, attention_mask: Tensor | None = None) -> Tensor:
        mask = (
            torch.ones_like(x, dtype=torch.bool).transpose(-2, -1)
            if attention_mask is None
            else attention_mask[:, :, 0]
        )
        x = x.transpose(-2, -1)
        for conv in self.convs:
            x = x * mask
            x = conv(x)
            if self.num_remove > 0:
                x = x[..., : -self.num_remove]
            x = self.layer_norm(x)
            x = F.gelu(x)
        x = x * mask
        return x.transpose(-2, -1)


class SelfAttention(nn.Module):
    """Multihead Self Attention module."""

    def __init__(self, embed_dim: int, num_heads: int, *, qkv_bias: bool, dropout: float = 0.0) -> None:
        super().__init__()
        if embed_dim % num_heads != 0:
            raise ValueError(f"embed_dim ({embed_dim}) must be divisible by num_heads ({num_heads})")

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.qkv = nn.Linear(embed_dim, embed_dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(embed_dim, embed_dim, bias=True)

    def forward(self, x: Tensor, attention_mask: Tensor | None = None) -> Tensor:
        batch, seq, dim = x.shape
        qkv = self.qkv(x)
        q, k, v = qkv.split(self.embed_dim, dim=2)
        k = k.view(batch, seq, self.num_heads, dim // self.num_heads).transpose(1, 2)
        q = q.view(batch, seq, self.num_heads, dim // self.num_heads).transpose(1, 2)
        v = v.view(batch, seq, self.num_heads, dim // self.num_heads).transpose(1, 2)
        dropout = self.dropout if self.training else 0.0
        x = F.scaled_dot_product_attention(q, k, v, attn_mask=attention_mask, dropout_p=dropout)
        x = x.transpose(1, 2).contiguous().view(batch, seq, dim)
        return self.proj(x)


class FeedForward(nn.Module):
    """Layer that follows attention layer in encoder layer."""

    def __init__(self, embed_dim: int, interm_features: int, interm_dropout: float) -> None:
        super().__init__()
        self.intermediate_dense = nn.Linear(embed_dim, interm_features)
        self.intermediate_dropout = nn.Dropout(interm_dropout)
        self.output_dense = nn.Linear(interm_features, embed_dim)

    def forward(self, x: Tensor) -> Tensor:
        x = self.intermediate_dense(x)
        x = F.gelu(x)
        x = self.intermediate_dropout(x)
        return self.output_dense(x)


class TransformerLayer(nn.Module):
    """Combines multihead self attention and feed forward."""

    def __init__(
        self, attention: SelfAttention, dropout: float, feed_forward: FeedForward, *, layer_norm_first: bool
    ) -> None:
        super().__init__()
        self.attention = attention
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(attention.embed_dim)
        self.layer_norm_first = layer_norm_first
        self.feed_forward = feed_forward
        self.final_layer_norm = nn.LayerNorm(attention.embed_dim)

    def forward(self, x: Tensor, attention_mask: Tensor | None = None) -> tuple[Tensor, Tensor | None]:
        residual = x
        if self.layer_norm_first:
            x = self.layer_norm(x)
        x = self.attention(x, attention_mask=attention_mask)
        x = self.dropout(x)
        x = residual + x
        if self.layer_norm_first:
            residual = x
            layer_result = self.feed_forward(self.final_layer_norm(x))
            x = self.dropout(layer_result)
            x = residual + x
        else:
            x = self.layer_norm(x)
            residual = x
            layer_result = self.feed_forward(x)
            x = self.dropout(layer_result)
            x = self.final_layer_norm(residual + x)
        return x, layer_result


class Transformer(nn.Module):
    """Transformer module with positional convolutional embeddings."""

    def __init__(
        self,
        layers: nn.ModuleList,
        pos_conv_embed: ConvPositionalEmbedding,
        dropout: float,
        layer_drop: float,
        *,
        layer_norm_first: bool,
    ) -> None:
        super().__init__()
        self.pos_conv_embed = pos_conv_embed
        self.layer_norm = nn.LayerNorm(pos_conv_embed.embed_dim)
        self.layer_norm_first = layer_norm_first
        self.layer_drop = layer_drop
        self.dropout = nn.Dropout(dropout)
        self.layers = layers
        self.apply(self.init_weights)

    @staticmethod
    def init_weights(module: nn.Module) -> None:
        if isinstance(module, SelfAttention):
            nn.init.xavier_uniform_(module.qkv.weight, gain=1 / math.sqrt(2))
            nn.init.xavier_uniform_(module.proj.weight)
            nn.init.constant_(module.proj.bias, 0.0)
        elif isinstance(module, nn.Linear):
            nn.init.normal_(module.weight, std=0.02)
            if module.bias is not None:
                nn.init.constant_(module.bias, 0)

    def forward(self, x: Tensor, attention_mask: Tensor | None = None) -> Tensor:
        x = x + self.pos_conv_embed(x, attention_mask)
        if self.layer_norm_first:
            x = self.layer_norm(x)
        x = self.dropout(x)
        for layer in self.layers:
            if not (self.training and torch.rand(1) <= self.layer_drop):
                x, _ = layer(x, attention_mask)
        if not self.layer_norm_first:
            x = self.layer_norm(x)
        return x

    def get_intermediate_outputs(
        self, x: Tensor, attention_mask: Tensor | None = None, num_layers: int | None = None
    ) -> list[Tensor]:
        if num_layers is not None and not 0 < num_layers <= len(self.layers):
            exception = f"`num_layers` must be between [1, {len(self.layers)}]"
            raise ValueError(exception)

        ret: list[Tensor] = []
        x = x + self.pos_conv_embed(x, attention_mask)
        if self.layer_norm_first:
            x = self.layer_norm(x)
        x = self.dropout(x)
        for layer in self.layers:
            x_output, layer_result = layer(x, attention_mask)
            if not (self.training and torch.rand(1) <= self.layer_drop):
                x = x_output.clone()
            ret.append(layer_result)
            if num_layers is not None and len(ret) >= num_layers:
                return ret
        return ret


class Codebook(nn.Module):
    def __init__(self, encoder_embed_dim: int, codebook_size: int, codebook_decay: float) -> None:
        super().__init__()
        self.codebook_decay = codebook_decay
        self.codebook_size = codebook_size
        codebook = torch.randn(encoder_embed_dim, codebook_size, dtype=torch.float32).unsqueeze(0)
        codebook = F.instance_norm(codebook).transpose(1, 2).squeeze().contiguous()
        self.codebook: Tensor
        self.counts: Tensor
        self.register_buffer("codebook", codebook)
        self.register_buffer("counts", torch.ones(codebook_size, dtype=torch.float32).contiguous())

    @torch.no_grad()
    def forward(self, target: Tensor) -> Tensor:
        codebook = self.codebook / self.counts.unsqueeze(1)
        labels = torch.cdist(target, codebook, p=2).argmin(1)
        onehot_target = F.one_hot(labels, self.codebook_size).float()
        if self.training:
            self.step(onehot_target, target)
        return onehot_target

    @torch.no_grad()
    def step(self, onehot_target: Tensor, target: Tensor) -> None:
        if self.codebook_decay >= 1.0:
            return
        count = onehot_target.sum(0)
        memory = torch.matmul(onehot_target.t(), target)
        if dist.is_initialized():
            dist.all_reduce(memory)  # Sum of embeddings
            dist.all_reduce(count)  # Total counts
        alpha = torch.ones_like(count).unsqueeze(1)
        alpha[count != 0] = self.codebook_decay
        self.counts = alpha.squeeze(1) * self.counts + (1 - alpha).squeeze(1) * count
        self.codebook = alpha * self.codebook + (1 - alpha) * memory


def get_components(
    cfg: DinoSRConfig,
) -> tuple[FeatureExtractor, FeatureProjection, Transformer, nn.ModuleList, nn.ModuleList]:
    if cfg.extractor_mode not in {"group_norm", "layer_norm"}:
        raise ValueError(cfg.extractor_mode)
    blocks = nn.ModuleList()
    in_channels = 1
    for i, (out_channels, kernel_size, stride) in enumerate(cfg.extractor_conv_layer_config):
        layer_norm: nn.Module | None = None
        if cfg.extractor_mode == "group_norm" and i == 0:
            layer_norm = nn.GroupNorm(num_groups=out_channels, num_channels=out_channels, affine=True)
        elif cfg.extractor_mode == "layer_norm":
            layer_norm = LayerNorm(normalized_shape=out_channels, elementwise_affine=True)
        blocks.append(
            ConvLayerBlock(in_channels, out_channels, kernel_size, stride, layer_norm, bias=cfg.extractor_conv_bias)
        )
        in_channels = out_channels
    feature_extractor = FeatureExtractor(blocks)

    feature_projection = FeatureProjection(cfg.extractor_conv_layer_config[-1][0], cfg.encoder_embed_dim)
    pos_conv = ConvPositionalEmbedding(
        cfg.encoder_embed_dim, cfg.encoder_pos_conv_kernel, cfg.encoder_pos_conv_groups, cfg.encoder_pos_conv_depth
    )
    layers = nn.ModuleList()
    for _ in range(cfg.encoder_num_layers):
        attention = SelfAttention(
            cfg.encoder_embed_dim,
            cfg.encoder_num_heads,
            qkv_bias=cfg.encoder_qkv_bias,
            dropout=cfg.encoder_attention_dropout,
        )
        feed_forward = FeedForward(
            cfg.encoder_embed_dim, cfg.encoder_ff_interm_features, cfg.encoder_ff_interm_dropout
        )
        layers.append(
            TransformerLayer(
                attention, cfg.encoder_dropout, feed_forward, layer_norm_first=cfg.encoder_layer_norm_first
            )
        )
    student = Transformer(
        layers,
        pos_conv,
        cfg.encoder_dropout,
        cfg.encoder_layer_drop,
        layer_norm_first=not cfg.encoder_layer_norm_first,
    )

    heads = nn.ModuleList(
        nn.Sequential(nn.Linear(cfg.encoder_embed_dim, cfg.codebook_size), nn.LogSoftmax(dim=-1))
        for _ in range(cfg.num_codebooks)
    )
    codebooks = nn.ModuleList(
        Codebook(cfg.encoder_embed_dim, cfg.codebook_size, cfg.codebook_decay) for _ in range(cfg.num_codebooks)
    )
    return feature_extractor, feature_projection, student, heads, codebooks



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""SpidR and DinoSR models."""

from spidr.models.dinosr import DinoSR
from spidr.models.spidr import SpidR
from spidr.models.utils import build_model, dinosr_base_original, dinosr_base_reproduced, spidr_base

__all__ = ["DinoSR", "SpidR", "build_model", "dinosr_base_original", "dinosr_base_reproduced", "spidr_base"]



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Speech dataset and batch sampler. Adapted from torchaudio."""

import abc
import math
from collections.abc import Iterator
from pathlib import Path
from typing import Literal

import torch
from torch import Tensor
from torch import distributed as dist
from torch.nn import functional as F
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import BatchSampler, DataLoader, Dataset, DistributedSampler
from torchcodec.decoders import AudioDecoder

from spidr.config import DEFAULT_CONV_LAYER_CONFIG, SAMPLE_RATE, DataConfig, MaskingConfig
from spidr.data.masks import MaskGenerator
from spidr.data.utils import bytes_from_archive, read_manifest


def verify_lengths(min_len: int, max_len: int, batch_size: int | None, max_token_count: int | None) -> None:
    exception = ""
    if not 0 <= min_len <= max_len:
        exception += "min_len must be less than or equal to max_len \n"
    if max_token_count is not None and batch_size is not None:
        exception += "max_token_count and batch_size cannot be set simultaneously \n"
    if max_token_count is None and batch_size is None:
        exception += "max_token_count or batch_size must be set \n"
    if max_token_count is not None and max_len > max_token_count:
        exception += "max_len must be less than or equal to max_token_count \n"
    if exception:
        raise ValueError(exception)


def get_buckets(lengths: list[int], num_buckets: int, uniform_limits: tuple[int, int] | None) -> dict[int, Tensor]:
    buckets: dict[int, list[int]] = {}
    if uniform_limits is not None:
        boundaries = torch.linspace(uniform_limits[0] - 1, uniform_limits[1] + 1, num_buckets + 1)
    else:
        boundaries = torch.quantile(
            torch.tensor(lengths, dtype=torch.float32),
            torch.linspace(0, 1, num_buckets + 1),
            interpolation="lower",
        )[1:]
    bucket_ids = torch.bucketize(torch.tensor(lengths), boundaries)
    for i in range(bucket_ids.size(0)):
        bucket_id = int(bucket_ids[i])
        if bucket_id in buckets:
            buckets[bucket_id].append(i)
        else:
            buckets[bucket_id] = [i]
    return dict(sorted([(k, torch.as_tensor(v, dtype=torch.int)) for k, v in buckets.items()]))


class BucketizeBatchSampler(BatchSampler):
    """Batch sampler that groups samples of similar length into buckets."""

    def __init__(
        self,
        *,
        lengths: list[int],
        num_buckets: int,
        min_len: int,
        max_len: int | None,
        max_token_count: int | None,
        batch_size: int | None,
        seed: int,
        bucket_method: Literal["uniform", "percentile"],
        shuffle: bool,
        drop_last: bool,
    ) -> None:
        if max_len is None:
            max_len = max(lengths)
        verify_lengths(min_len, max_len, batch_size, max_token_count)
        filtered_length_idx = [(min(length, max_len), i) for i, length in enumerate(lengths) if min_len <= length]
        if not filtered_length_idx:
            exception = "No samples with length in the range"
            raise ValueError(exception)

        sorted_filtered_length_idx = sorted(filtered_length_idx, key=lambda x: x[0])
        self.lengths = [e[0] for e in sorted_filtered_length_idx]
        self.indices = [e[1] for e in sorted_filtered_length_idx]
        self.max_token_count = max_token_count
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed
        if self.shuffle:
            self.generator = torch.Generator().manual_seed(self.seed)
        self.drop_last = drop_last
        if bucket_method == "uniform":
            uniform_limits = (min_len, max_len)
        elif bucket_method == "percentile":
            uniform_limits = None
        else:
            raise ValueError("bucket_method must be either 'uniform' or 'percentile'")
        self.buckets = get_buckets(self.lengths, num_buckets, uniform_limits)
        self._update_iter_list()

    def _update_iter_list(self) -> None:
        if self.shuffle:
            for k in self.buckets:
                new_idx = torch.randperm(self.buckets[k].size(0), generator=self.generator)
                self.buckets[k] = self.buckets[k][new_idx]
        self.iter_list = []
        total_len, batch = 0, []
        max_batch_size = self.max_token_count or self.batch_size
        for k in self.buckets:
            for i in range(self.buckets[k].size(0)):
                index = int(self.buckets[k][i])
                sample_length = self.lengths[index] if self.max_token_count else 1
                if total_len + sample_length <= max_batch_size:
                    batch.append(self.indices[index])
                    total_len += sample_length
                else:
                    self.iter_list.append(batch)
                    batch = [self.indices[index]]
                    total_len = sample_length
        if len(batch) > 0 and (self.max_token_count or not self.drop_last):
            self.iter_list.append(batch)

    def set_epoch(self, epoch: int) -> None:
        self.seed += epoch
        self.generator.manual_seed(self.seed)
        self._update_iter_list()

    def __iter__(self) -> Iterator[list[int]]:
        return iter(self.iter_list)

    def __len__(self) -> int:
        return len(self.iter_list)


class DistributedBatchSampler(DistributedSampler):
    """Distributed sampler for BucketizeBatchSampler."""

    def __init__(self, batch_sampler: BucketizeBatchSampler, *, seed: int, shuffle: bool, drop_last: bool) -> None:
        self.batch_sampler = batch_sampler
        self.num_replicas = dist.get_world_size() if dist.is_initialized() else 1
        self.rank = dist.get_rank() if dist.is_initialized() else 0
        self.shuffle = shuffle
        self.epoch = 0
        self.seed = seed
        self.drop_last = drop_last
        self.shuffle = shuffle
        indices = self.batch_sampler.iter_list
        if self.drop_last and len(indices) % self.num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when using this Sampler.
            self.num_samples = math.ceil((len(indices) - self.num_replicas) / self.num_replicas)
        else:
            self.num_samples = math.ceil(len(indices) / self.num_replicas)

    def set_epoch(self, epoch: int) -> None:
        self.batch_sampler.set_epoch(epoch)
        return super().set_epoch(epoch)

    def __iter__(self) -> Iterator[list[int]]:
        if self.shuffle:
            generator = torch.Generator().manual_seed(self.seed + self.epoch)
            perm = torch.randperm(len(self.batch_sampler.iter_list), generator=generator).tolist()
            indices = [self.batch_sampler.iter_list[i] for i in perm]
        else:
            indices = self.batch_sampler.iter_list
        if self.drop_last:
            self.total_size = len(indices) - len(indices) % self.num_replicas
        else:
            padding_size = self.num_replicas - len(indices) % self.num_replicas
            indices += indices[:padding_size]
            self.total_size = len(indices)
        self.num_samples = self.total_size // self.num_replicas
        self.subset = indices[self.rank : self.total_size : self.num_replicas]
        if len(self.subset) != self.num_samples:
            exception = f"Rank {self.rank} has subset of length {len(self.subset)} but expected {self.num_samples}"
            raise ValueError(exception)
        return iter(self.subset)

    def __len__(self) -> int:
        return self.num_samples


class SpeechDataset(Dataset, abc.ABC):
    """Dataset to load chunks of audio files."""

    def __init__(self, manifest_path: Path | str, *, normalize: bool) -> None:
        super().__init__()
        self.manifest = read_manifest(manifest_path)
        self.normalize = normalize

    def __len__(self) -> int:
        return len(self.manifest)

    @abc.abstractmethod
    def _load_audio(self, index: int) -> tuple[Tensor, int]:
        pass

    def __getitem__(self, index: int) -> Tensor:
        waveform, sr = self._load_audio(index)
        if sr != SAMPLE_RATE or waveform.shape[0] != 1:
            raise ValueError(index)
        if self.normalize:
            waveform = F.layer_norm(waveform, waveform.shape)
        return waveform.squeeze()


class SpeechDatasetFromArchive(SpeechDataset):
    def _load_audio(self, index: int) -> tuple[Tensor, int]:
        entry = self.manifest[index].to_dicts()[0]
        data = bytes_from_archive(entry["archive"], entry["byte_offset"], entry["byte_size"])
        samples = AudioDecoder(data).get_all_samples()
        return samples.data, samples.sample_rate


class SpeechDatasetFromFiles(SpeechDataset):
    def _load_audio(self, index: int) -> tuple[Tensor, int]:
        samples = AudioDecoder(self.manifest[index, "path"]).get_all_samples()
        return samples.data, samples.sample_rate


def speech_dataset(manifest_path: Path | str, *, normalize: bool) -> SpeechDataset:
    with Path(manifest_path).open("r", encoding="utf-8") as f:
        columns = set(f.readline().strip().split(","))
    if {"fileid", "path", "num_samples", "archive", "byte_offset", "byte_size"}.issubset(columns):
        return SpeechDatasetFromArchive(manifest_path, normalize=normalize)
    return SpeechDatasetFromFiles(manifest_path, normalize=normalize)


def conv_length(shapes: list[tuple[int, int, int]], length: Tensor) -> Tensor:
    for _, kernel_size, stride in shapes:
        length = torch.div(length - kernel_size, stride, rounding_mode="floor") + 1
        length = torch.max(torch.zeros_like(length), length)
    return length


def crop_audio(waveform: Tensor, num_samples: int, max_sample_size: int, *, rand_crop: bool) -> tuple[Tensor, int]:
    frame_offset = 0
    length = waveform.size(0)
    num_samples = min(num_samples, max_sample_size)
    if length > num_samples and rand_crop:
        frame_offset = int(torch.randint(length - num_samples, size=(1,)))
    elif length < num_samples:
        num_samples = length
    return waveform[frame_offset : frame_offset + num_samples], num_samples


class SpeechCollatorWithMasking:
    def __init__(
        self,
        mask_generator: MaskGenerator,
        *,
        max_sample_size: int,
        conv_layer_config: list[tuple[int, int, int]],
        enable_padding: bool,
        rand_crop: bool,
    ) -> None:
        self.mask_generator = mask_generator
        self.max_sample_size = max_sample_size
        self.conv_layer_config = conv_layer_config
        self.enable_padding = enable_padding
        self.rand_crop = rand_crop

    def __call__(self, batch: list[Tensor]) -> tuple[Tensor, Tensor, Tensor | None]:
        num_samples = max(len(wav) for wav in batch) if self.enable_padding else min(len(wav) for wav in batch)
        wavs_with_len = [crop_audio(wav, num_samples, self.max_sample_size, rand_crop=self.rand_crop) for wav in batch]
        wav_list, wav_lengths = zip(*wavs_with_len, strict=True)
        wavs = pad_sequence(wav_list, batch_first=True)
        lengths = conv_length(self.conv_layer_config, torch.tensor(wav_lengths))
        batch_size, max_len = wavs.size(0), int(lengths.max())
        padding_mask = torch.arange(max_len, device=lengths.device).expand(batch_size, max_len) >= lengths[:, None]
        attn_mask = ~padding_mask[:, None, None, :].expand(batch_size, 1, max_len, max_len)
        mask_indices = self.mask_generator(padding_mask)[0]
        return wavs, attn_mask, mask_indices


def build_dataloader(
    data_cfg: DataConfig,
    mask_cfg: MaskingConfig,
    *,
    conv_layer_config: list[tuple[int, int, int]] | None = None,
) -> DataLoader:
    dataset = speech_dataset(data_cfg.manifest, normalize=data_cfg.normalize)
    batch_sampler = DistributedBatchSampler(
        BucketizeBatchSampler(
            lengths=dataset.manifest["num_samples"].to_list(),
            num_buckets=data_cfg.num_buckets,
            min_len=data_cfg.min_sample_size,
            max_len=data_cfg.max_sample_size,
            max_token_count=data_cfg.max_batch_length,
            batch_size=None,
            seed=data_cfg.random_seed,
            bucket_method=data_cfg.bucket_method,
            shuffle=True,
            drop_last=data_cfg.drop_last,
        ),
        seed=data_cfg.random_seed,
        shuffle=True,
        drop_last=data_cfg.drop_last,
    )
    collate_fn = SpeechCollatorWithMasking(
        MaskGenerator(mask_cfg),
        max_sample_size=data_cfg.max_sample_size,
        conv_layer_config=conv_layer_config or DEFAULT_CONV_LAYER_CONFIG,
        enable_padding=data_cfg.enable_padding,
        rand_crop=data_cfg.rand_crop,
    )
    return DataLoader(
        dataset,
        batch_sampler=batch_sampler,
        num_workers=data_cfg.num_workers,
        collate_fn=collate_fn,
        pin_memory=data_cfg.pin_memory,
        prefetch_factor=data_cfg.prefetch_factor,
        persistent_workers=data_cfg.persistent_workers,
        generator=torch.Generator().manual_seed(
            data_cfg.random_seed + (dist.get_rank() if dist.is_initialized() else 0)
        ),
    )



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Entry point for launching training runs."""

import argparse
import os
from pathlib import Path

from spidr.config import read_config
from spidr.slurm import default_train_job_config, launch_with_submitit, slurm_config_parse_args
from spidr.tools import init_logger
from spidr.train import train

if __name__ == "__main__":
    cluster = default_train_job_config(os.getenv("SPIDR_CLUSTER"))
    parser = argparse.ArgumentParser(
        description="Launch training runs.",
        allow_abbrev=False,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("configs", type=Path, nargs="+", help="TOML config file(s) for the training run(s).")
    parser.add_argument("-A", "--account", type=str, required=True, help="Slurm account")
    parser.add_argument("-N", "--nodes", type=int, required=True, help="Number of nodes")
    parser.add_argument("-G", "--gpus-per-node", type=int, required=True, help="GPUs per node")
    parser.add_argument("-c", "--cpus-per-task", type=int, help="CPUs per task", default=cluster.get("cpus_per_task"))
    parser.add_argument("--mem-per-gpu", type=str, help="Memory per GPU", default=cluster.get("mem_per_gpu"))
    parser.add_argument("-t", "--time", type=int, help="Time limit in minutes", default=cluster.get("time"))
    parser.add_argument("-C", "--constraint", type=str, help="Slurm constraint", default=cluster.get("constraint"))
    parser.add_argument("-p", "--partition", type=str, help="Slurm partition", default=cluster.get("partition"))
    parser.add_argument("-q", "--qos", type=str, help="Slurm QoS", default=cluster.get("qos"))
    parser.add_argument("--dump", type=Path, help="Submitit dump", default=os.getenv("SPIDR_SUBMITIT_FOLDER"))
    args = parser.parse_args()

    if args.dump is None:
        parser.error("Dump directory must be specified with --dump or by setting $SPIDR_SUBMITIT_FOLDER")
    init_logger()
    jobs = [(train, (read_config(cfg),)) for cfg in args.configs]
    name = jobs[0][1][0].run.wandb_name
    launch_with_submitit(name, jobs, args.dump, slurm_config_parse_args(args))



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""DinoSR model definition."""

import copy
from collections.abc import Iterable
from functools import partial

import torch
from torch import Tensor, nn
from torch.nn import functional as F

from spidr.config import DinoSRConfig
from spidr.models.components import Transformer, get_components
from spidr.models.metrics import perplexity


def ema_scheduler(step: int, start_decay: float, final_decay: float, final_step: int, freeze_step: int) -> float:
    if step < final_step:
        pct = 1 - step / final_step
        return final_decay - (final_decay - start_decay) * pct
    if step < freeze_step:
        return final_decay
    return 1


def init_teacher(
    student: Transformer, exclude_layers: Iterable[str], *, init_weights: bool = True
) -> tuple[Transformer, set[str]]:
    teacher = copy.deepcopy(student).float()
    if init_weights:
        teacher.apply(teacher.init_weights)
    teacher.eval()
    teacher.requires_grad_(requires_grad=False)
    teacher_exclude_layers: set[str] = set()
    for name, param in teacher.named_parameters():
        param.detach_()
        if any(name.startswith(ex) for ex in exclude_layers):
            teacher_exclude_layers.add(name)
    return teacher, teacher_exclude_layers


class DinoSR(nn.Module):
    def __init__(self, cfg: DinoSRConfig | None = None) -> None:
        super().__init__()
        if cfg is None:
            cfg = DinoSRConfig()

        self.feature_extractor, self.feature_projection, self.student, self.heads, self.codebooks = get_components(cfg)
        self.teacher, self.teacher_exclude_layers = init_teacher(self.student, cfg.ema_exclude_layers)
        self.ema_scheduler = partial(
            ema_scheduler,
            start_decay=cfg.ema_start_decay,
            final_decay=cfg.ema_final_decay,
            final_step=cfg.ema_final_step,
            freeze_step=cfg.freeze_step,
        )
        self.projection_dropout = nn.Dropout(cfg.encoder_projection_dropout)
        self.freeze_step = cfg.freeze_step
        self._extractor_frozen = False
        self.mask_embedding = nn.Parameter(torch.FloatTensor(cfg.encoder_embed_dim))
        nn.init.uniform_(self.mask_embedding)
        self.current_step = nn.Buffer(torch.zeros(1, dtype=torch.int64))

    def train(self, mode: bool = True) -> "DinoSR":  # noqa: FBT001, FBT002
        super().train(mode)
        self.teacher.eval()
        return self

    @property
    def num_codebooks(self) -> int:
        return len(self.codebooks)

    def freeze_extractor(self) -> None:
        for p in self.feature_extractor.parameters():
            p.requires_grad = False
        for p in self.feature_projection.parameters():
            p.requires_grad = False
        self._extractor_frozen = True

    @torch.no_grad()
    def _inner_ema(self, decay: torch.Tensor) -> None:
        for (ema_n, ema_p), model_p in zip(self.teacher.named_parameters(), self.student.parameters(), strict=True):
            if ema_n in self.teacher_exclude_layers:
                ema_p.copy_(model_p)
            else:
                ema_p.lerp_(model_p, 1 - decay)
        for ema_b, model_b in zip(self.teacher.buffers(), self.student.buffers(), strict=True):
            ema_b.copy_(model_b)

    def update_ema(self, step: int) -> float:
        self.current_step.fill_(step)
        decay = self.ema_scheduler(step)
        if not self._extractor_frozen and step >= self.freeze_step:
            self.freeze_extractor()
        if 0.0 < decay < 1.0:
            self._inner_ema(torch.tensor(decay, device=self.current_step.device))
        return decay

    def get_intermediate_outputs(self, waveforms: Tensor, *, attention_mask: Tensor | None = None) -> list[Tensor]:
        x = self.feature_extractor(waveforms)
        x = self.feature_projection(x)
        return self.student.get_intermediate_outputs(x, attention_mask)

    def get_codebooks(
        self,
        waveform: Tensor,
        *,
        attention_mask: Tensor | None = None,
        onehot: bool = False,
    ) -> list[Tensor | None]:
        x = self.feature_extractor(waveform)
        x = self.feature_projection(x)
        x = self.student(x, attention_mask)
        codebooks: list[Tensor | None] = [None] * (len(self.student.layers) - self.num_codebooks)
        for i in range(self.num_codebooks):
            codebook = self.heads[i](x).float().exp().squeeze()
            if onehot:
                codebook = F.one_hot(codebook.argmax(dim=-1), codebook.size(-1))
            codebooks.append(codebook)
        return codebooks

    def forward(
        self, waveforms: Tensor, *, mask: Tensor | None = None, attention_mask: Tensor | None = None
    ) -> tuple[Tensor, dict[str, Tensor]]:
        feats = self.feature_extractor(waveforms)
        feats = self.feature_projection(feats)
        x = feats.clone()
        x = self.projection_dropout(x)
        if mask is not None:
            x = torch.where(mask.unsqueeze(-1), self.mask_embedding.to(x.dtype).expand_as(x), x)
        else:
            mask = torch.ones((x.shape[0], x.shape[1]), dtype=torch.bool, device=x.device)
        x = self.student(x, attention_mask)
        mask_indices = torch.nonzero(mask, as_tuple=True)
        x = x[mask_indices]

        with torch.no_grad():
            targets = self.teacher.get_intermediate_outputs(feats, attention_mask)[-self.num_codebooks :]
            targets = [F.instance_norm(tl.float().transpose(1, 2)).transpose(1, 2)[mask_indices] for tl in targets]

        losses = torch.zeros(x.shape[0], device=x.device)
        target_ppl, pred_ppl = torch.zeros(1, device=x.device), torch.zeros(1, device=x.device)
        for i, target in enumerate(targets):
            log_pred = self.heads[i](x)
            onehot_target = self.codebooks[i](target)
            target_ppl += perplexity(onehot_target)
            pred_ppl += perplexity(log_pred.exp())
            losses += torch.sum(-onehot_target * log_pred, dim=-1)

        return (losses / self.num_codebooks), {
            "target_ppl": target_ppl / self.num_codebooks,
            "pred_ppl": pred_ppl / self.num_codebooks,
        }



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Data utilities."""

import mmap
import os
from pathlib import Path

import polars as pl
from torchcodec.decoders import AudioDecoder


def num_samples(source: str | Path | bytes, *, verify: bool = False) -> int:
    metadata = AudioDecoder(source).metadata
    samples = metadata.duration_seconds_from_header * metadata.sample_rate
    if verify and not samples.is_integer():
        raise ValueError(
            f"Number of samples {samples} is not an integer"
            + (f" in {source}" if isinstance(source, (str, Path)) else "")
        )
    return int(samples)


def bytes_from_archive(archive: Path | str, offset: int, file_size: int) -> bytes:
    with Path(archive).open("rb") as path, mmap.mmap(path.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_o:
        return mmap_o[offset : offset + file_size]


def read_manifest(path: Path | str) -> pl.DataFrame:
    path = Path(path)
    if path.suffix == ".csv":
        return pl.read_csv(path)
    if path.suffix != ".tsv":
        raise ValueError("Only .csv and .tsv files are supported")
    with path.open("r") as file:
        root = Path(file.readline().strip())
    if not root.is_dir():
        raise ValueError("First line must be the root directory of the dataset")
    return (
        pl.scan_csv(path, separator="\t", skip_rows=1, has_header=False, new_columns=["fileid", "num_samples"])
        .with_columns((f"{root}{os.sep}" + pl.col("fileid")).alias("path"))
        .select("fileid", "path", "num_samples")
        .collect()
    )



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Various tools for training."""

from spidr.tools.git import write_git_info_if_available
from spidr.tools.logger import init_logger
from spidr.tools.meters import AverageMeters
from spidr.tools.metrics_queue import MetricsQueue
from spidr.tools.profiler import profiler_context

__all__ = ["AverageMeters", "MetricsQueue", "init_logger", "profiler_context", "write_git_info_if_available"]



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Metrics."""

import itertools
import math
from collections.abc import Iterable, Sequence
from itertools import groupby
from pathlib import Path
from typing import NamedTuple

import torch
import torch.compiler
from torch import Tensor
from torch import distributed as dist


@torch.no_grad()
@torch.compiler.disable
def perplexity(y: Tensor, *, tokenwise: bool = False) -> Tensor:
    if tokenwise:
        y = torch.exp2((-y * (y + 1e-8).log2()).sum(-1))
    if dist.is_initialized():
        n = torch.tensor([y.size(0)], device=y.device)
        y = y.sum(0)
        dist.all_reduce(n)
        dist.all_reduce(y)
        y /= n
    else:
        y = y.mean(0)
    if not tokenwise:
        y = torch.exp2((-y * (y + 1e-8).log2()).sum())
    return y


class NonFiniteError(Exception):
    """Non finite values were found."""


def params_norm(
    params: torch.Tensor | Iterable[torch.Tensor], norm_type: float = 2.0, *, error_if_nonfinite: bool = False
) -> torch.Tensor:
    params = (params,) if isinstance(params, torch.Tensor) else tuple(params)
    total_norm = torch.linalg.norm(torch.stack(torch._foreach_norm(params, norm_type)), norm_type)
    if error_if_nonfinite and torch.logical_or(total_norm.isnan(), total_norm.isinf()):
        raise NonFiniteError
    return total_norm


PhonesAndUnits = tuple[list[str], list[int]]


def build_count_matrix(
    data: dict[str | Path, PhonesAndUnits],
    num_units: int,
    num_phones: int,
    repeat: int,
    max_length_diff: int = 2,
) -> tuple[torch.Tensor, list[str]]:
    phone_to_index, idx, phone_indices, unit_indices = {}, 0, [], []
    for path, phones_and_units in data.items():
        phones = phones_and_units[0]
        units = list(itertools.chain.from_iterable(itertools.repeat(unit, repeat) for unit in phones_and_units[1]))
        if abs(len(phones) - len(units)) > max_length_diff:
            raise ValueError(
                f"More than {max_length_diff} frames of difference "
                f" after repetition ({len(phones)} vs {len(units)}) in {path}"
            )
        for phone, unit in zip(phones, units, strict=False):
            if phone not in phone_to_index:
                phone_to_index[phone] = idx
                idx += 1
            if phone_to_index[phone] >= num_phones or unit >= num_units:
                raise IndexError(
                    f"Phone '{phone}' or code '{unit}' is out of bounds. "
                    f"Adjust num_phones ({num_phones}) or num_units ({num_units}) "
                )
            phone_indices.append(phone_to_index[phone])
            unit_indices.append(unit)

    flattened_indices = torch.tensor(phone_indices) * num_units + torch.tensor(unit_indices)
    count_flat = torch.bincount(flattened_indices, minlength=num_phones * num_units)
    count = count_flat.reshape(num_phones, num_units)
    most_frequent_phones = torch.argsort(count.sum(dim=1), descending=True)
    phone_order = [{v: k for k, v in phone_to_index.items()}[idx.item()] for idx in most_frequent_phones]
    count = count[most_frequent_phones]
    return count, phone_order


class UnitsQuality(NamedTuple):
    phone_purity: float
    cluster_purity: float
    pnmi: float
    codebook_perplexity: float
    active_codewords: int
    bitrate: float


def bitrate(*, num_units: int, sequences: Iterable[Sequence[int]] | None = None, frame_step: int = 20) -> float:
    """Bitrate in bits per second.

    :param num_units: Number of distinct units. Equivalent to the codebook size.
    :param sequences: Sequences of units.
        If provided, deduplicate consecutive units in each sequence.
        If None, assume that one unit is produced every `frame_step` ms.
    :param frame_step: Frame step in milliseconds.
    """
    if sequences is None:
        return math.log2(num_units) * 1000 / frame_step
    total_deduplicated_units, total_seconds = 0, 0
    for units in sequences:
        total_seconds += frame_step * len(units) / 1000  # frame_step is in ms
        total_deduplicated_units += len(list(groupby(units)))
    num_units_per_second = total_deduplicated_units / total_seconds
    return math.log2(num_units) * num_units_per_second


def units_quality(
    data: dict[str | Path, PhonesAndUnits],
    *,
    num_units: int,
    num_phones: int,
    frame_step: int = 20,
    alignment_step: int = 10,
    eps: float = 1e-10,
) -> UnitsQuality:
    count = build_count_matrix(data, num_units, num_phones, frame_step // alignment_step)[0]
    proba = count / count.sum()
    phone_purity = proba.max(dim=0).values.sum()
    cluster_purity = proba.max(dim=1).values.sum()
    px = proba.sum(dim=1, keepdim=True)
    py = proba.sum(dim=0, keepdim=True)
    mutual_info = (proba * torch.log(proba / (px @ py + eps) + eps)).sum()
    entropy_x = (-px * torch.log(px + eps)).sum()
    pnmi = mutual_info / entropy_x
    proba_sum = count.sum(dim=0) / count.sum()
    codebook_perplexity = torch.exp2(-torch.sum(proba_sum * torch.log2(proba_sum + eps)))
    active_codewords = int((count.sum(dim=0) > 0).sum())
    return UnitsQuality(
        phone_purity=phone_purity.item(),
        cluster_purity=cluster_purity.item(),
        pnmi=pnmi.item(),
        codebook_perplexity=codebook_perplexity.item(),
        active_codewords=active_codewords,
        bitrate=bitrate(num_units=num_units, sequences=[units for (_, units) in data.values()], frame_step=frame_step),
    )


def proba_phone_code(
    data: dict[str | Path, PhonesAndUnits],
    *,
    num_units: int,
    num_phones: int,
    only_active: bool,
    frame_step: int = 20,
    alignment_step: int = 10,
) -> tuple[torch.Tensor, list[str], torch.Tensor]:
    count, phone_order = build_count_matrix(data, num_units, num_phones, frame_step // alignment_step)
    count_by_unit = count.sum(dim=0, keepdim=True)
    proba = torch.where(count_by_unit != 0, count / count_by_unit, torch.zeros_like(count))
    codes_order, argmax = [], proba.argmax(dim=0)
    for phn_idx in range(len(count)):
        indices = torch.where(argmax == phn_idx)[0]
        codes_order.extend(indices[torch.argsort(proba[phn_idx, indices], descending=True)])
    if only_active:
        codes_order = [code for code in codes_order if count[:, code].sum() > 0]
    return proba[:, codes_order], phone_order, torch.tensor(codes_order)



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Run validation with an existing checkpoint."""

import json
import logging

import torch
from torch.utils.data import DataLoader

from spidr.config import Config, ResumeConfig
from spidr.data import build_dataloader
from spidr.environment import set_seed, setup_environment, setup_pytorch
from spidr.models import DinoSR, build_model
from spidr.tools import init_logger

logger = logging.getLogger()


@torch.no_grad()
def validate(model: DinoSR, loader: DataLoader, device: torch.device) -> dict[str, float]:
    model.eval()
    total_loss = torch.zeros(1, device=device)
    total_target_ppl = torch.zeros(1, device=device)
    total_pred_ppl = torch.zeros(1, device=device)
    for waveforms, attn_mask, mask in loader:
        loss, outputs = model(waveforms.to(device), mask=mask.to(device), attention_mask=attn_mask.to(device))
        total_loss += loss.mean()
        total_target_ppl += outputs["target_ppl"]
        total_pred_ppl += outputs["pred_ppl"]
    total_loss /= len(loader)
    total_target_ppl /= len(loader)
    total_pred_ppl /= len(loader)
    return {"loss": total_loss.item(), "target_ppl": total_target_ppl.item(), "pred_ppl": total_pred_ppl.item()}


def validate_existing_checkpoint(cfg: Config, resume: ResumeConfig) -> None:
    logger.info("Starting validation for step %s", resume.step)
    init_logger()
    set_seed(cfg.run.random_seed)
    setup_pytorch(use_deterministic=cfg.run.use_deterministic)
    setup_environment()
    if resume.checkpoint is None:
        logger.error("No checkpoint found, validation failed")
        return
    logger.info("Loading model from %s", resume.checkpoint)
    model = build_model(model_type=cfg.run.model_type, checkpoint=resume.checkpoint).eval().to("cuda")
    for val_name, val_cfg in cfg.validation.items():
        logger.info("Validating on %s", val_name)
        loader = build_dataloader(val_cfg, cfg.masking, conv_layer_config=cfg.model.extractor_conv_layer_config)
        val_output = {"step": resume.step, "group": val_name} | validate(model, loader, torch.device("cuda"))
        logger.info("Validation on %s done: loss %s", val_name, val_output["loss"])
        with resume.results.open("a") as f:
            f.write(json.dumps(val_output) + "\n")



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Model loading utilities."""

import json
import typing as tp
import warnings
from collections import OrderedDict
from itertools import product
from pathlib import Path
from typing import Any

import torch
from torch.hub import load_state_dict_from_url
from torch.nn.modules.utils import consume_prefix_in_state_dict_if_present

from spidr.config import DinoSRConfig, ModelType, SpidRConfig
from spidr.models.dinosr import DinoSR
from spidr.models.spidr import SpidR


def model_from_raw_checkpoint(
    model_class: type[SpidR] | type[DinoSR],
    config_class: type[DinoSRConfig] | type[SpidRConfig],
    ckpt: str | Path,
) -> DinoSR | SpidR:
    path = Path(ckpt)
    if path.suffix != ".pt":
        raise ValueError("Only .pt files are supported.")
    if (path.parent / "config.json").is_file():
        with (path.parent / "config.json").open() as f:
            json_cfg = json.load(f)
            if "model" in json_cfg:
                json_cfg = json_cfg["model"]
            cfg = config_class(**json_cfg)
    else:
        warnings.warn("Config file not found when loading checkpoint. Using default config.", stacklevel=2)
        cfg = config_class()
    instance = model_class(cfg)
    state_dict = torch.load(path, map_location="cpu", weights_only=True)
    if "model" in state_dict:
        state_dict = state_dict["model"]
    consume_prefix_in_state_dict_if_present(state_dict, "module.")
    instance.load_state_dict(state_dict)
    return instance


def build_model(
    *,
    cfg: DinoSRConfig | SpidRConfig | None = None,
    model_type: ModelType | None = None,
    checkpoint: str | Path | None = None,
) -> DinoSR | SpidR:
    if checkpoint is not None:
        match model_type:
            case "dinosr":
                return model_from_raw_checkpoint(DinoSR, DinoSRConfig, checkpoint)
            case "spidr":
                return model_from_raw_checkpoint(SpidR, SpidRConfig, checkpoint)
            case _:
                raise ValueError(f"Model type not recognized, acceptable models are {tp.get_args(ModelType)}.")
    if isinstance(cfg, SpidRConfig):
        return SpidR(cfg)
    if isinstance(cfg, DinoSRConfig):  # Must be last condition because SpidRConfig is subclass of DinoSRConfig
        return DinoSR(cfg)
    raise ValueError("Invalid cfg class")


def state_dict_from_dinosr_fairseq_checkpoint(checkpoint: dict[str, Any]) -> OrderedDict[str, torch.Tensor]:
    mapping = {
        "mask_emb": "mask_embedding",
        "layer_norm.bias": "feature_projection.layer_norm.bias",
        "layer_norm.weight": "feature_projection.layer_norm.weight",
        "post_extract_proj.bias": "feature_projection.projection.bias",
        "post_extract_proj.weight": "feature_projection.projection.weight",
        "encoder.layer_norm.bias": "student.layer_norm.bias",
        "encoder.layer_norm.weight": "student.layer_norm.weight",
    }
    for i in range(7):
        mapping |= {
            f"feature_extractor.conv_layers.{i}.0.weight": f"feature_extractor.conv_layers.{i}.conv.weight",
            f"feature_extractor.conv_layers.{i}.2.1.bias": f"feature_extractor.conv_layers.{i}.layer_norm.bias",
            f"feature_extractor.conv_layers.{i}.2.1.weight": f"feature_extractor.conv_layers.{i}.layer_norm.weight",
        }
    for i in range(5):
        mapping |= {
            f"encoder.pos_conv.{i}.0.bias": f"student.pos_conv_embed.convs.{i}.bias",
            f"encoder.pos_conv.{i}.0.weight": f"student.pos_conv_embed.convs.{i}.weight",
        }
    for i in range(8):
        mapping |= {
            f"_codebook{i}": f"codebooks.{i}.codebook",
            f"_codebook_cnts{i}": f"codebooks.{i}.counts",
            f"heads.{i}.bias": f"heads.{i}.0.bias",
            f"heads.{i}.weight": f"heads.{i}.0.weight",
        }
    for i in range(12):
        fairseq, mine = "encoder.layers", "student.layers"
        mapping |= {
            f"{fairseq}.{i}.fc1.bias": f"{mine}.{i}.feed_forward.intermediate_dense.bias",
            f"{fairseq}.{i}.fc1.weight": f"{mine}.{i}.feed_forward.intermediate_dense.weight",
            f"{fairseq}.{i}.fc2.bias": f"{mine}.{i}.feed_forward.output_dense.bias",
            f"{fairseq}.{i}.fc2.weight": f"{mine}.{i}.feed_forward.output_dense.weight",
            f"{fairseq}.{i}.final_layer_norm.bias": f"{mine}.{i}.final_layer_norm.bias",
            f"{fairseq}.{i}.final_layer_norm.weight": f"{mine}.{i}.final_layer_norm.weight",
            f"{fairseq}.{i}.self_attn.k_proj.bias": f"{mine}.{i}.attention.k_proj.bias",
            f"{fairseq}.{i}.self_attn.k_proj.weight": f"{mine}.{i}.attention.k_proj.weight",
            f"{fairseq}.{i}.self_attn.out_proj.bias": f"{mine}.{i}.attention.proj.bias",
            f"{fairseq}.{i}.self_attn.out_proj.weight": f"{mine}.{i}.attention.proj.weight",
            f"{fairseq}.{i}.self_attn.q_proj.bias": f"{mine}.{i}.attention.q_proj.bias",
            f"{fairseq}.{i}.self_attn.q_proj.weight": f"{mine}.{i}.attention.q_proj.weight",
            f"{fairseq}.{i}.self_attn.v_proj.bias": f"{mine}.{i}.attention.v_proj.bias",
            f"{fairseq}.{i}.self_attn.v_proj.weight": f"{mine}.{i}.attention.v_proj.weight",
            f"{fairseq}.{i}.self_attn_layer_norm.bias": f"{mine}.{i}.layer_norm.bias",
            f"{fairseq}.{i}.self_attn_layer_norm.weight": f"{mine}.{i}.layer_norm.weight",
        }

    ema_mapping = {
        "layer_norm.bias": "teacher.layer_norm.bias",
        "layer_norm.weight": "teacher.layer_norm.weight",
    }
    for i in range(5):
        ema_mapping |= {
            f"pos_conv.{i}.0.bias": f"teacher.pos_conv_embed.convs.{i}.bias",
            f"pos_conv.{i}.0.weight": f"teacher.pos_conv_embed.convs.{i}.weight",
        }
    for i in range(12):
        fairseq, mine = "layers", "teacher.layers"
        ema_mapping |= {
            f"{fairseq}.{i}.fc1.bias": f"{mine}.{i}.feed_forward.intermediate_dense.bias",
            f"{fairseq}.{i}.fc1.weight": f"{mine}.{i}.feed_forward.intermediate_dense.weight",
            f"{fairseq}.{i}.fc2.bias": f"{mine}.{i}.feed_forward.output_dense.bias",
            f"{fairseq}.{i}.fc2.weight": f"{mine}.{i}.feed_forward.output_dense.weight",
            f"{fairseq}.{i}.final_layer_norm.bias": f"{mine}.{i}.final_layer_norm.bias",
            f"{fairseq}.{i}.final_layer_norm.weight": f"{mine}.{i}.final_layer_norm.weight",
            f"{fairseq}.{i}.self_attn.k_proj.bias": f"{mine}.{i}.attention.k_proj.bias",
            f"{fairseq}.{i}.self_attn.k_proj.weight": f"{mine}.{i}.attention.k_proj.weight",
            f"{fairseq}.{i}.self_attn.out_proj.bias": f"{mine}.{i}.attention.proj.bias",
            f"{fairseq}.{i}.self_attn.out_proj.weight": f"{mine}.{i}.attention.proj.weight",
            f"{fairseq}.{i}.self_attn.q_proj.bias": f"{mine}.{i}.attention.q_proj.bias",
            f"{fairseq}.{i}.self_attn.q_proj.weight": f"{mine}.{i}.attention.q_proj.weight",
            f"{fairseq}.{i}.self_attn.v_proj.bias": f"{mine}.{i}.attention.v_proj.bias",
            f"{fairseq}.{i}.self_attn.v_proj.weight": f"{mine}.{i}.attention.v_proj.weight",
            f"{fairseq}.{i}.self_attn_layer_norm.bias": f"{mine}.{i}.layer_norm.bias",
            f"{fairseq}.{i}.self_attn_layer_norm.weight": f"{mine}.{i}.layer_norm.weight",
        }
    new_state_dict = OrderedDict({mapping[k]: v for k, v in checkpoint["model"].items() if k != "_ema"})
    new_state_dict |= {ema_mapping[k]: v for k, v in checkpoint["model"]["_ema"].items()}
    keys_to_remove = []
    for m, i in product(["student.layers", "teacher.layers"], range(12)):
        new_state_dict[f"{m}.{i}.attention.qkv.weight"] = torch.cat(
            [
                new_state_dict[f"{m}.{i}.attention.q_proj.weight"],
                new_state_dict[f"{m}.{i}.attention.k_proj.weight"],
                new_state_dict[f"{m}.{i}.attention.v_proj.weight"],
            ],
            dim=0,
        )
        new_state_dict[f"{m}.{i}.attention.qkv.bias"] = torch.cat(
            [
                new_state_dict[f"{m}.{i}.attention.q_proj.bias"],
                new_state_dict[f"{m}.{i}.attention.k_proj.bias"],
                new_state_dict[f"{m}.{i}.attention.v_proj.bias"],
            ]
        )
        keys_to_remove += [
            f"{m}.{i}.attention.q_proj.weight",
            f"{m}.{i}.attention.k_proj.weight",
            f"{m}.{i}.attention.v_proj.weight",
            f"{m}.{i}.attention.q_proj.bias",
            f"{m}.{i}.attention.k_proj.bias",
            f"{m}.{i}.attention.v_proj.bias",
        ]
    for k in keys_to_remove:
        del new_state_dict[k]
    try:
        current_step = checkpoint["last_optimizer_state"]["state"][0]["step"]
    except KeyError:
        current_step = 0
    new_state_dict["current_step"] = torch.tensor([current_step])
    return new_state_dict


def spidr_base(*, pretrained: bool = True, check_hash: bool = False, progress: bool = True) -> SpidR:
    model = SpidR(SpidRConfig())
    if pretrained:
        url = "https://dl.fbaipublicfiles.com/shared/devai/assets/models/spidr_base.pt"
        checkpoint = load_state_dict_from_url(url, check_hash=check_hash, progress=progress, map_location="cpu")
        model.load_state_dict(checkpoint["model"])
    model.eval()
    return model


def dinosr_base_reproduced(*, pretrained: bool = True, check_hash: bool = False, progress: bool = True) -> DinoSR:
    model = DinoSR(DinoSRConfig())
    if pretrained:
        url = "https://dl.fbaipublicfiles.com/shared/devai/assets/models/dinosr_base_reproduced.pt"
        checkpoint = load_state_dict_from_url(url, check_hash=check_hash, progress=progress, map_location="cpu")
        model.load_state_dict(checkpoint["model"])
    model.eval()
    return model


def dinosr_base_original(*, pretrained: bool = True, check_hash: bool = False, progress: bool = True) -> DinoSR:
    model = DinoSR(DinoSRConfig(encoder_qkv_bias=True))
    if pretrained:
        url = "https://data.csail.mit.edu/placesaudio/dinosr/dinosr.ckpt"
        state_dict = load_state_dict_from_url(url, check_hash=check_hash, progress=progress)
        state_dict = state_dict_from_dinosr_fairseq_checkpoint(state_dict)
        model.load_state_dict(state_dict)
    model.eval()
    return model



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Logging utilities."""

import logging
import sys
from pathlib import Path
from typing import ClassVar


class ColoredFormatter(logging.Formatter):
    COLORS: ClassVar[dict[str, str]] = {
        "NOTSET": "\033[0m",  # No color
        "DEBUG": "\033[36m",  # Cyan
        "INFO": "\033[32m",  # Green
        "WARNING": "\033[33m",  # Yellow
        "ERROR": "\033[31m",  # Red
        "CRITICAL": "\033[35m",  # Magenta
    }
    RESET = "\033[0m"  # Reset color code

    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)
        return f"{self.COLORS[record.levelname]}{message}{self.RESET}"


def init_logger(file: str | Path | None = None, *, with_colors: bool = False) -> None:
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    template = "[spidr] %(asctime)s - %(levelname)s - %(name)s - %(message)s"
    formatter = (ColoredFormatter if with_colors else logging.Formatter)(template)

    stdout = logging.StreamHandler(sys.stdout)
    stdout.setLevel(logging.DEBUG)
    stdout.setFormatter(formatter)
    stderr = logging.StreamHandler(sys.stderr)
    stderr.setLevel(logging.WARNING)
    stderr.setFormatter(formatter)
    logger.handlers = []
    logger.addHandler(stdout)
    logger.addHandler(stderr)

    if file is not None:
        handler = logging.FileHandler(file, "a")
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        logger.addHandler(handler)



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Launch Slurm jobs with submitit."""

import argparse
import importlib.resources
import json
import logging
import os
import subprocess
from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path
from typing import ParamSpec, TypeVar

import submitit
from submitit.core.utils import FailedJobError

from spidr.config import SlurmConfig
from spidr.tools import write_git_info_if_available

logger = logging.getLogger()


def current_codebase_path() -> Path:
    import spidr  # noqa: PLC0415

    return Path(spidr.__file__).parent.parent


def copy_code_to_submitit_folder(output_dir: Path) -> None:
    output_dir.mkdir(parents=True, exist_ok=True)
    root = Path(importlib.resources.files(__package__)).parent.parent
    logger.info("Copying: %s to: %s", root, output_dir)
    rsync_cmd = (
        f"rsync -ar --copy-links "
        f"--exclude='.git/' "
        f"--exclude './data/' "
        f"--exclude './paper/' "
        f"--exclude __pycache__ "
        f"--exclude '*.pyc' "
        f"--exclude '*.ipynb' "
        f"--exclude '*.err' "
        f"--exclude '*.out' "
        f"--exclude '*.log' "
        f"--exclude '*.pt' "
        f"--exclude '.ruff_cache/' "
        f"--exclude '.mypy_cache/' "
        f"{root}/ {output_dir}"
    )
    subprocess.call([rsync_cmd], shell=True)
    logger.info("Copy done.")


T = TypeVar("T")
P = ParamSpec("P")


def make_checkpointable(fn: Callable[P, T]) -> Callable[P, T]:  # noqa: UP047
    # Disable linting because "Only parameter specification variables defined in global scope can be pickled."
    def checkpoint(*args: P.args, **kwargs: P.kwargs) -> submitit.helpers.DelayedSubmission:
        return submitit.helpers.DelayedSubmission(fn, args, kwargs)

    fn.checkpoint = checkpoint  # ty: ignore[unresolved-attribute]
    return fn


def default_train_job_config(cluster_name: str | None) -> dict:
    match cluster_name:
        case None:
            return {}
        case "JZ_V100":
            return {"time": 1200, "cpus_per_task": 10, "mem_per_gpu": None, "constraint": None}
        case "JZ_A100":
            return {"time": 1200, "cpus_per_task": 8, "mem_per_gpu": None, "constraint": "a100"}
        case "JZ_H100":
            return {"time": 1200, "cpus_per_task": 24, "mem_per_gpu": None, "constraint": "h100"}
        case "FAIR_AWS_A100":
            return {"time": 4320, "cpus_per_task": 12, "mem_per_gpu": "80g", "constraint": None}
    raise ValueError(f"Unknown cluster {cluster_name}")


def validation_job_config() -> SlurmConfig:
    try:
        account, cpus_per_task = os.environ["SLURM_JOB_ACCOUNT"], os.environ["SLURM_CPUS_PER_TASK"]
    except KeyError as error:
        raise ValueError("This function should only be used inside a SLURM job.") from error
    return SlurmConfig(account=account, nodes=1, gpus_per_node=1, time=60, cpus_per_task=int(cpus_per_task))


def slurm_config_parse_args(args: argparse.Namespace) -> SlurmConfig:
    config = SlurmConfig(account=args.account, nodes=args.nodes, gpus_per_node=args.gpus_per_node)
    for key in SlurmConfig.__optional_keys__:
        if hasattr(args, key) and getattr(args, key) is not None:
            config[key] = getattr(args, key)  # ty: ignore[invalid-key]
    return config


@submitit.helpers.clean_env()
def launch_with_submitit(
    job_name: str,
    jobs: list[tuple[Callable, tuple]],
    dump: str | Path,
    slurm: SlurmConfig,
    *,
    copy_code: bool = True,
) -> bool:
    launch_time = datetime.now(UTC).strftime("%Y%m%d-%H%M%S")
    folder = Path(dump) / job_name
    if folder.is_dir():
        folder = Path(dump) / f"{job_name}_{launch_time}"
    folder.mkdir(parents=True)

    if copy_code:
        copy_code_to_submitit_folder(folder / "code")
        codebase_path = folder / "code" / "src"
    else:  # No-op in most situations, but correctly handles nested cases
        codebase_path = current_codebase_path()

    write_git_info_if_available(folder)
    job_info = {"job_name": job_name, "launch_time": launch_time} | slurm
    (folder / "job-info.json").write_text(json.dumps(job_info, indent=2))

    os.chdir(codebase_path)
    logger.info("Launching from %s", codebase_path)
    executor = submitit.SlurmExecutor(folder=folder / "%j")
    executor.update_parameters(job_name=job_name, ntasks_per_node=slurm["gpus_per_node"], **slurm)
    try:
        with executor.batch():
            for fn, args in jobs:
                executor.submit(make_checkpointable(fn), *args)
    except FailedJobError as error:
        logger.warning("Didn't managed to submit %s: %s", job_name, error)
        return False
    logger.info("Submitted %s (%s jobs)", job_name, len(jobs))
    return True



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""PyTorch profiler."""

import contextlib
import importlib.resources
import string
import tempfile
from collections.abc import Generator
from functools import partial
from pathlib import Path

from torch.profiler import ProfilerAction, profile


def html_trace_handler(p: profile, path: str | Path) -> None:
    """Adapted from https://github.com/facebookresearch/lingua/blob/main/lingua/profiling.py."""
    with tempfile.NamedTemporaryFile(mode="w+b", suffix=".json") as fp:
        p.export_chrome_trace(fp.name)
        content = Path(fp.name).read_text(encoding="utf-8")
    viztracer = importlib.resources.files("viztracer")
    template = (viztracer / "html/trace_viewer_embedder.html").read_text(encoding="utf-8")
    sub = {"trace_viewer_full": (viztracer / "html/trace_viewer_full.html").read_text(encoding="utf-8")}
    sub["json_data"] = content.replace("</script>", "<\\/script>")
    Path(path).write_text(string.Template(template).substitute(sub), encoding="utf-8")


def scheduler_fn(step: int, skip_first: int, warmup: int, active: int) -> ProfilerAction:
    if step < skip_first or step >= skip_first + warmup + active:
        return ProfilerAction.NONE
    if step < skip_first + warmup:
        return ProfilerAction.WARMUP
    if step < skip_first + warmup + active - 1:
        return ProfilerAction.RECORD
    return ProfilerAction.RECORD_AND_SAVE


class NullProfiler:
    """A no-op profiler that does nothing."""

    def step(self) -> None:  # noqa: PLR6301
        return


@contextlib.contextmanager
def profiler_context(
    path: str | Path | None,
    skip_first: int = 1_000,
    warmup: int = 5,
    active: int = 2,
) -> Generator[profile | NullProfiler]:
    if path is not None:
        with profile(
            schedule=partial(scheduler_fn, skip_first=skip_first, warmup=warmup, active=active),
            on_trace_ready=partial(html_trace_handler, path=path),
            record_shapes=True,
            profile_memory=True,
            with_stack=False,
            with_flops=True,
        ) as profiler:
            yield profiler
    else:
        yield NullProfiler()



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Mask creation."""

import torch
from torch import Tensor, nn

from spidr.config import MaskingConfig, MaskingStrategy


def mask_lengths(
    num_mask: int, mask_length: int, mask_other: float, mask_type: MaskingStrategy, generator: torch.Generator | None
) -> torch.Tensor:
    match mask_type:
        case "static":
            return torch.full((num_mask,), mask_length)
        case "uniform":
            return torch.randint(int(mask_other), mask_length * 2 + 1, size=(num_mask,), generator=generator)
        case "normal":
            lengths = torch.normal(mask_length, mask_other, size=(num_mask,), generator=generator)
            return torch.maximum(torch.ones(1), torch.round(lengths)).int()
        case "poisson":
            lengths = torch.poisson(torch.rand(num_mask) * mask_length, generator=generator)
            return torch.round(lengths).int()
    raise ValueError(f"unknown mask strategy: {mask_type}")


def mask_id_overlap(lengths: torch.Tensor, num_mask: int, sz: int, generator: torch.Generator | None) -> torch.Tensor:
    min_len = min(lengths)
    if sz - min_len <= num_mask:
        min_len = sz - num_mask - 1
    mask_idc = torch.randperm(sz - min_len, generator=generator)[:num_mask]
    return torch.tensor([mask_idc[j] + offset for j in range(len(mask_idc)) for offset in range(lengths[j])])


def mask_id_no_overlap(
    lengths: torch.Tensor, min_space: int, sz: int, generator: torch.Generator | None
) -> torch.Tensor:
    mask_idc_list: list[int] = []

    def arrange(mask_idc_list: list[int], s: int, e: int, length: int, keep_length: int) -> list[tuple[int, int]]:
        span_start = int(torch.randint(s, e - length, size=(1,), generator=generator))
        mask_idc_list.extend(span_start + i for i in range(length))

        new_parts = []
        if span_start - s - min_space >= keep_length:
            new_parts.append((s, span_start - min_space + 1))
        if e - span_start - keep_length - min_space > keep_length:
            new_parts.append((span_start + length + min_space, e))
        return new_parts

    parts = [(0, sz)]
    min_length = min(lengths)
    for length in sorted(lengths, reverse=True):
        lens = torch.tensor([e - s for s, e in parts], dtype=torch.int)
        lens[lens < length + min_space] = 0
        l_sum = lens.sum()
        if l_sum == 0:
            break
        probs = lens / l_sum
        c = torch.distributions.Categorical(probs).sample()  # WARNING: does not accept `generator`
        s, e = parts.pop(c)
        parts.extend(arrange(mask_idc_list, s, e, length, min_length))
    return torch.tensor(mask_idc_list)


def compute_mask_indices(
    shape: tuple[int, int],
    padding_mask: Tensor | None,
    mask_prob: float,
    mask_length: int,
    mask_type: MaskingStrategy,
    mask_other: float,
    min_masks: int,
    min_space: int,
    *,
    no_overlap: bool,
    generator: torch.Generator | None,
) -> Tensor:
    """Compute random mask spans for a given shape."""
    batch_size, frame = shape
    mask = torch.full((batch_size, frame), fill_value=False)
    # add a random number for probabilistic rounding
    all_num_mask = int(mask_prob * frame / float(mask_length) + torch.rand(1, generator=generator))
    all_num_mask = max(min_masks, all_num_mask)

    mask_idcs = []
    for i in range(batch_size):
        if padding_mask is not None:
            sz = frame - padding_mask[i].long().sum().item()
            # add a random number for probabilistic rounding
            num_mask = int(mask_prob * sz / float(mask_length) + torch.rand(1, generator=generator))
            num_mask = max(min_masks, num_mask)
        else:
            sz, num_mask = frame, all_num_mask
        lengths = mask_lengths(num_mask, mask_length, mask_other, mask_type, generator)
        if sum(lengths) == 0:
            lengths[0] = min(mask_length, sz - 1)
        mask_idc = (
            mask_id_no_overlap(lengths, min_space, sz, generator)
            if no_overlap
            else mask_id_overlap(lengths, num_mask, sz, generator)
        )
        mask_idcs.append(torch.unique(mask_idc[mask_idc < sz]))

    min_len = min(len(m) for m in mask_idcs)
    for i, mask_idc in enumerate(mask_idcs):
        if len(mask_idc) > min_len:
            idx = torch.randperm(len(mask_idc), generator=generator)[:min_len].long()
            mask[i, mask_idc[idx]] = True
        else:
            mask[i, mask_idc] = True
    return mask


class MaskGenerator(nn.Module):
    """Generate the masks for masked prediction."""

    def __init__(self, config: MaskingConfig) -> None:
        super().__init__()
        self.config = config

    def forward(
        self,
        padding_mask: Tensor,
        *,
        channels: int | None = None,
        generator: torch.Generator | None = None,
    ) -> tuple[Tensor | None, Tensor | None]:
        batch, time = padding_mask.shape
        if self.config.mask_prob > 0:
            mask_indices = compute_mask_indices(
                (batch, time),
                padding_mask,
                self.config.mask_prob,
                self.config.mask_length,
                self.config.mask_selection,
                self.config.mask_other,
                min_masks=2,
                min_space=self.config.mask_min_space,
                no_overlap=self.config.no_mask_overlap,
                generator=generator,
            )
        else:
            mask_indices = None
        if self.config.mask_channel_prob > 0:
            if channels is None:
                raise ValueError("Must set 'channels' to mask channel-wise")
            mask_channel_indices = compute_mask_indices(
                (batch, channels),
                None,
                self.config.mask_channel_prob,
                self.config.mask_channel_length,
                self.config.mask_channel_selection,
                self.config.mask_channel_other,
                min_masks=0,
                min_space=self.config.mask_channel_min_space,
                no_overlap=self.config.no_mask_channel_overlap,
                generator=generator,
            )
            mask_channel_indices = mask_channel_indices.unsqueeze(1).expand(-1, time, -1)
        else:
            mask_channel_indices = None
        return mask_indices, mask_channel_indices



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Watch for metrics added to a JSONL file and yield them in order.

This whole snippet is a workaround for wandb's limitations.
We want to train the model in a distributed setting, and launch validation
jobs in separate processes. The outcome of these validation jobs should be
available in the same wandb run as the training job. However, wandb does not
support multiple processes opened that can write to the same run.
It also only supports having a monotonic increasing step number.
Therefore, we write the metrics from the validation jobs in a JSONL file like this:
```
{"step": 100, "group": "dev-clean", "metric1": 0.1, "metric2": 0.2}
{"step": 100, "group": "dev-other", "metric3": 0.2, "metric4": 0.3}
{"step": 200, "group": "dev-clean", "metric1": 0.3, "metric2": 0.4}
```
The `MetricsQueue` class reads this file, and yields the metrics in order of
increasing step. It also waits for the next step to be available before
yielding the metrics for the following steps: this is useful if a job for
step N+1 has ended before the job for step N.
This approach has limitations:
- If some lines are removed or modified in the JSONL file, the `MetricsQueue`
  will raise a ValueError.
- If a job fails, the metrics for the following steps will be available in the
  JSONL but will no be logged.
  You can run `python -m spidr.tools.metrics_queue /path/to/jsonl --runid <runid> --entity <entity>`
  to push the unlogged metrics to wandb when training has ended.
"""

import difflib
import json
from collections import defaultdict
from collections.abc import Generator, Sequence
from pathlib import Path
from typing import Any


def pull_inserted_lines(a: Sequence[str], b: Sequence[str]) -> Generator[str, None, None]:
    """Adapted from difflib.Differ.compare.

    a and b are lines from two files. a is the old version of the file, b is the new one.
    This function will yield the lines that were inserted in b compared to a.
    Any line that is either removed or modified will raise a ValueError.
    """
    cruncher = difflib.SequenceMatcher(None, a, b)
    for tag, _, _, blo, bhi in cruncher.get_opcodes():
        if tag == "equal":
            continue
        if tag == "insert":
            for i in range(blo, bhi):
                yield b[i]
        else:
            raise ValueError(
                f"Invalid tag {tag}, only 'equal' and 'insert' are supported."
                f"This mean that some lines were either removed or modified."
            )


class MetricsQueue:
    """Wait and yield metrics from a JSONL file."""

    def __init__(self, path: Path, interval: int, start: int) -> None:
        self.interval = interval
        self.path = Path(path)
        self._queues: dict[str, list[tuple[int, dict[str, Any]]]] = {}
        self._next_step: defaultdict[str, int] = defaultdict(lambda: start)
        self._previous: list[str] = []
        if self.path.is_file():
            self.reset()

    def reset(self) -> None:
        """Reset the expected next step."""
        with self.path.open() as f:
            self._previous = f.read().splitlines()

        for line in self.path.read_text().splitlines():
            metric = json.loads(line)
            step = metric.pop("step")
            group = metric.pop("group")
            self._next_step[group] = max(self._next_step[group], step + self.interval)

    def pull(self) -> Generator[str, None, None]:
        """Pull the new lines from the JSONL file and yield them."""
        if not self.path.is_file():
            return
        with self.path.open() as f:
            lines = f.read().splitlines()
        yield from pull_inserted_lines(self._previous, lines)
        self._previous = lines

    def populate_queues(self) -> None:
        """Add the new metrics to the queues for each group of metrics."""
        for line in self.pull():
            metric: dict[str, Any] = json.loads(line)
            step: int = metric.pop("step")
            group: str = metric.pop("group")
            if group not in self._queues:
                self._queues[group] = []
            self._queues[group].append((step, metric))

    def __call__(self) -> Generator[dict[str, Any], None, None]:
        """Yield the metrics to log in order of increasing step for each group."""
        self.populate_queues()
        new_queues = {}
        for group, queue in self._queues.items():
            idx = 0
            sorted_queue = sorted(queue, key=lambda x: x[0])
            for step, metric in sorted_queue:
                if step < self._next_step[group]:
                    raise ValueError(
                        f"This should not be happening: a new entry from an existing group ({group}) has been added, "
                        f"and its step ({step}) is lower than the next expected step ({self._next_step[group]}). "
                        f"The wrong entry is '{metric}'."
                    )
                if step == self._next_step[group]:
                    self._next_step[group] += self.interval
                    idx += 1
                    yield {f"{group}/{name}": value for name, value in metric.items()} | {f"{group}/step": step}
                else:
                    break
            new_queues[group] = sorted_queue[idx:]
        self._queues = new_queues


def push_later_to_wandb(path: Path, runid: str | None, entity: str) -> None:
    """Push unlogged metrics from a finished run to wandb."""
    import polars as pl  # noqa: PLC0415
    import wandb  # noqa: PLC0415

    metrics = pl.read_ndjson(path)
    api = wandb.Api()
    with (path.parent / "config.json").open() as f:
        config = json.load(f)
    project, name = config["run"]["wandb_project"], config["run"]["wandb_name"]

    if runid is None:
        candidates = [run for run in api.runs(f"{entity}/{project}") if run.name == name and run.state != "running"]
        if len(candidates) > 1:
            raise ValueError(f"Found {len(candidates)} candidate runs. Please specify the run ID.")
        if not candidates:
            raise ValueError("No candidate run found: wrong name or still running.")
        runid = candidates[0].id
    history = api.run(f"{entity}/{project}/{runid}").history()
    max_steps = {group: history[f"{group}/step"].max().astype(int) for group in metrics["group"].unique()}
    metrics = metrics.filter(pl.col("step") > pl.col("group").replace_strict(max_steps))
    if len(metrics) == 0:
        return
    wandb.init(id=runid, project=project, entity=entity, resume="must", dir=Path(config["run"]["workdir"]).resolve())
    for _, row in metrics.sort(by=["step", "group"]).iter_rows():
        wandb.log({f"{row['group']}/{name}": value for name, value in row.items() if name != "group"})
    wandb.finish()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Push metrics that were not logged to wandb.")
    parser.add_argument("path", type=Path, help="Path to the JSONL file.")
    parser.add_argument("--entity", type=str, required=True, help="Wandb account.")
    parser.add_argument("--runid", type=str, help="Run ID in wandb.")
    args = parser.parse_args()
    push_later_to_wandb(args.path, args.runid, args.entity)



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Retrieve and write git information."""

import importlib.resources
import subprocess
from pathlib import Path
from typing import NamedTuple


class GitInfo(NamedTuple):
    branch: str
    commit: str
    diff: str


def git_available() -> bool:
    git_cmd = ["git", "rev-parse", "--is-inside-work-tree"]
    try:
        return subprocess.run(git_cmd, capture_output=True, check=False, text=True).returncode == 0
    except FileNotFoundError:
        return False


def git_info() -> GitInfo:
    root = Path(importlib.resources.files(__package__)).parent.parent
    git_cmd = ["git", "-C", root, "rev-parse", "HEAD", "--abbrev-ref", "HEAD"]
    info = subprocess.run(git_cmd, capture_output=True, check=True, text=True)
    commit, branch = info.stdout.strip().split("\n")
    git_diff_cmd = ["git", "-C", root, "diff"]
    diff = subprocess.run(git_diff_cmd, capture_output=True, check=True, text=True)
    return GitInfo(branch=branch, commit=commit, diff=diff.stdout)


def write_git_info_if_available(path: str | Path) -> None:
    if not git_available():
        return
    info = git_info()
    (Path(path) / "git-info.txt").write_text(f"branch: {info.branch}\ncommit: {info.commit}\n")
    (Path(path) / "git-diff.patch").write_text(info.diff)



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Training loop."""

import dataclasses
import json
import logging
import os
from contextlib import ExitStack
from pathlib import Path

import torch
import wandb
from torch import distributed as dist
from torch.nn.parallel import DistributedDataParallel
from torch.nn.utils import clip_grad_norm_
from tqdm import tqdm

from spidr.checkpoint import Checkpointer, remove_param_group
from spidr.config import Config, ResumeConfig, read_config
from spidr.data import build_dataloader
from spidr.environment import setup_training
from spidr.models import build_model
from spidr.optimizer import build_optimizer
from spidr.slurm import launch_with_submitit, validation_job_config
from spidr.tools import AverageMeters, profiler_context
from spidr.validate import validate_existing_checkpoint

logger = logging.getLogger()


def init_wandb(cfg: Config) -> str | None:
    logger.debug("Initializing wandb")
    cfg.run.dir.mkdir(parents=True, exist_ok=True)
    (cfg.run.dir / "config.json").write_text(json.dumps(dataclasses.asdict(cfg), indent=4))
    run = wandb.init(
        project=cfg.run.wandb_project,
        name=cfg.run.wandb_name,
        mode=cfg.run.wandb_mode,
        dir=Path(cfg.run.workdir).resolve(),
        config=dataclasses.asdict(cfg),
    )
    wandb.define_metric("train/step")
    wandb.define_metric("train/*", step_metric="train/step")
    for prefix in cfg.validation:
        wandb.define_metric(f"{prefix}/step")
        wandb.define_metric(f"{prefix}/*", step_metric=f"{prefix}/step")
    return run.id


def launch_validation(cfg: Config, resume: ResumeConfig) -> None:
    if resume.checkpoint is None:
        logger.info("No checkpoint found, skipping validation at step %s", resume.step)
        return
    if cfg.run.slurm_validation is None and "SLURM_JOB_ID" not in os.environ:
        logger.info("Not training in a Slurm job and no slurm_validation config provided, skipping validation")
        return
    logger.info("Launching validation at step %s", resume.step)
    launch_with_submitit(
        f"val-{resume.step}",
        [(validate_existing_checkpoint, (cfg, resume))],
        cfg.run.dir / "validation",
        cfg.run.slurm_validation or validation_job_config(),
        copy_code=False,
    )


def train(cfg: Config) -> None:  # noqa: PLR0914, PLR0915, C901
    with ExitStack() as stack:
        logger.info("Starting job")
        setup_training(cfg.run.random_seed, use_deterministic=cfg.run.use_deterministic)
        stack.callback(dist.destroy_process_group)
        global_rank, world_size = dist.get_rank(), dist.get_world_size()
        is_main = global_rank == 0
        if is_main:
            init_wandb(cfg)
            stack.callback(wandb.finish)

        logger.info("Building model, optimizer, and dataloaders")
        device = torch.device(f"cuda:{os.environ['LOCAL_RANK']}")
        dtype = {"float32": torch.float32, "float16": torch.float16, "bfloat16": torch.bfloat16}[cfg.optimizer.dtype]
        model = build_model(cfg=cfg.model, model_type=cfg.run.model_type, checkpoint=cfg.run.init_ckpt)
        model = model.to(device).train()
        optimizer, scaler, scheduler = build_optimizer(model, cfg.optimizer)
        loader = build_dataloader(cfg.data, cfg.masking, conv_layer_config=cfg.model.extractor_conv_layer_config)
        dist.barrier(device_ids=[device.index])
        ckpt = Checkpointer(cfg.run.dir, cfg.run.save_interval, cfg.run.keep_latest)
        ckpt.init_state(model=model, optimizer=optimizer, scheduler=scheduler, scaler=scaler)
        resuming = ckpt.load_existing_run()
        step, epoch = int(ckpt.step), int(ckpt.epoch)
        if not resuming and is_main and ckpt.save(step, epoch):
            launch_validation(cfg, ResumeConfig(step=step, checkpoint=ckpt.last, results=ckpt.metrics))
        if cfg.run.compile:
            model.compile(dynamic=True)
            model._inner_ema = torch.compile(model._inner_ema)
        ddp_model = DistributedDataParallel(model, device_ids=[device.index], find_unused_parameters=True)

        logger.info("Starting training loop")
        meters = AverageMeters(["loss", "grad_norm", "batch_size", "target_ppl", "pred_ppl"], device=device)
        profiler = stack.enter_context(profiler_context(cfg.run.dir / "trace.html" if is_main else None))
        pbar = stack.enter_context(tqdm(total=cfg.optimizer.max_steps, initial=step, disable=not is_main))
        while step < cfg.optimizer.max_steps:
            epoch += 1
            loader.batch_sampler.set_epoch(epoch)
            logger.info("Starting epoch %s", epoch)
            for waveforms, attn_mask, mask in loader:
                if step >= cfg.optimizer.max_steps:
                    break
                if step == cfg.model.freeze_step and len(optimizer.param_groups) > 1:
                    remove_param_group(optimizer, 1)

                with torch.autocast("cuda", dtype, cfg.optimizer.mixed_precision):
                    loss, outputs = ddp_model(
                        waveforms.to(device),
                        mask=mask.to(device),
                        attention_mask=attn_mask.to(device),
                    )
                num_frames = torch.tensor(loss.size(0), dtype=torch.long, device=device)
                dist.all_reduce(num_frames)
                loss = loss.sum() * world_size / num_frames
                scaler.scale(loss).backward()
                scaler.unscale_(optimizer)
                grad_norm = clip_grad_norm_(ddp_model.parameters(), cfg.optimizer.max_norm)
                scaler.step(optimizer)
                scaler.update()
                optimizer.zero_grad(set_to_none=True)
                lr = scheduler.get_last_lr()[0]
                scheduler.step()
                step += 1
                ema_decay = model.update_ema(step)

                meters.update(loss=loss.detach(), batch_size=waveforms.size(0), grad_norm=grad_norm)
                meters.update(target_ppl=outputs["target_ppl"], pred_ppl=outputs["pred_ppl"])
                pbar.update()
                if is_main and step % cfg.run.log_interval == 0:
                    infos = meters.pop() | {"lr": lr, "ema_decay": ema_decay * 1000, "step": step, "epoch": epoch}
                    wandb.log({f"train/{key}": value for key, value in infos.items()})
                    pbar.set_postfix(loss=infos["loss"], target_ppl=infos["target_ppl"], pred_ppl=infos["pred_ppl"])
                if is_main and ckpt.save(step, epoch):
                    launch_validation(cfg, ResumeConfig(step=step, checkpoint=ckpt.last, results=ckpt.metrics))
                    for val_metric in ckpt.find_new_metrics():
                        wandb.log(val_metric)
                profiler.step()

        if is_main and ckpt.save_final(step, epoch):
            launch_validation(cfg, ResumeConfig(step=step, checkpoint=ckpt.last, results=ckpt.metrics))
        logger.info("Training finished")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("config", type=str)
    cfg = read_config(parser.parse_args().config)
    train(cfg)



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Keep track of metrics."""

import torch


class AverageMeter:
    """Compute and store the average and current value."""

    def __init__(self, device: torch.device, dtype: torch.dtype = torch.float32) -> None:
        self.device = device
        self.dtype = dtype
        self.reset()

    def reset(self) -> None:
        self.avg = torch.zeros((), device=self.device, dtype=self.dtype)
        self.sum = torch.zeros((), device=self.device, dtype=self.dtype)
        self.count = torch.zeros((), device=self.device, dtype=torch.long)

    def update(self, val: torch.Tensor, n: int = 1) -> None:
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def pop(self) -> float:
        avg = self.avg
        self.reset()
        return avg.item()


class AverageMeters:
    """Record average meters for multiple values."""

    def __init__(self, names: list[str], device: torch.device, *, dtype: torch.dtype = torch.float32) -> None:
        self.names = names
        self.meters = {name: AverageMeter(device, dtype) for name in names}

    def __getitem__(self, name: str) -> AverageMeter:
        return self.meters[name]

    def reset(self) -> None:
        for meter in self.meters.values():
            meter.reset()

    def update(self, **kwargs: torch.Tensor) -> None:
        for name, value in kwargs.items():
            self.meters[name].update(value)

    def pop(self) -> dict[str, float]:
        return {name: meter.pop() for name, meter in self.meters.items()}



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Configuration dataclasses."""

import json
import tomllib
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, NotRequired, TypedDict

SAMPLE_RATE: int = 16_000
DEFAULT_CONV_LAYER_CONFIG: list[tuple[int, int, int]] = [(512, 10, 5)] + [(512, 3, 2)] * 4 + [(512, 2, 2)] * 2

MaskingStrategy = Literal["static", "uniform", "normal", "poisson"]
ModelType = Literal["dinosr", "spidr"]


class SlurmConfig(TypedDict):
    account: str
    nodes: int
    gpus_per_node: int
    qos: NotRequired[str]
    time: NotRequired[int]
    cpus_per_task: NotRequired[int]
    mem_per_gpu: NotRequired[str]
    constraint: NotRequired[str]
    partition: NotRequired[str]


@dataclass(frozen=True)
class RunConfig:
    """Training run configuration."""

    workdir: str
    wandb_project: str
    wandb_name: str
    wandb_mode: Literal["online", "offline", "disabled"] = "offline"
    random_seed: int = 0
    use_deterministic: bool = False
    log_interval: int = 200
    save_interval: int = 2_500
    init_ckpt: str | None = None
    compile: bool = True
    keep_latest: int = -1
    model_type: ModelType = "dinosr"
    slurm_validation: SlurmConfig | None = None

    @property
    def dir(self) -> Path:
        return Path(self.workdir).resolve() / self.wandb_project / self.wandb_name


@dataclass(frozen=True)
class DataConfig:
    """Dataloading and data sampling configuration."""

    manifest: str
    enable_padding: bool = False
    rand_crop: bool = True
    normalize: bool = True
    num_buckets: int = 1000
    min_sample_size: int = 2_000
    max_sample_size: int = 320_000
    max_batch_length: int = 3_800_000
    drop_last: bool = False
    num_workers: int = 10
    pin_memory: bool = False
    prefetch_factor: int = 2
    persistent_workers: bool = True
    random_seed: int = 0
    bucket_method: Literal["uniform", "percentile"] = "uniform"
    alignments_path: Path | None = None


@dataclass(frozen=True)
class OptimizerConfig:
    """AdamW and learning rate scheduler configuration."""

    lr: float = 5e-4
    betas: tuple[float, float] = (0.9, 0.95)
    weight_decay: float = 0.01
    eps: float = 1e-6
    max_norm: float = 10.0
    dtype: Literal["float32", "float16", "bfloat16"] = "bfloat16"

    init_lr_scale: float = 0.01
    final_lr_scale: float = 0.01
    max_steps: int = 400_000

    warmup_steps: int = 12_000
    hold_steps: int = 188_000
    decay_steps: int = 200_000
    rsqrt_timescale: int = 10_000
    rsqrt_shift: int = 0

    scheduler: Literal["tristage", "cosine", "rsqrt", "constant"] = "tristage"

    exclude_from_optimizer: list[str] = field(default_factory=lambda: ["teacher"])
    to_freeze: list[str] = field(default_factory=lambda: ["feature_extractor", "feature_projection"])

    @property
    def mixed_precision(self) -> bool:
        return self.dtype != "float32"


@dataclass(frozen=True)
class DinoSRConfig:
    """DinoSR configuration. Corresponding names in `fairseq` in comments."""

    extractor_mode: Literal["group_norm", "layer_norm"] = "layer_norm"  # model.extractor_mode
    extractor_conv_bias: bool = False  # model.conv_bias
    # model.conv_layer_config
    extractor_conv_layer_config: list[tuple[int, int, int]] = field(default_factory=lambda: DEFAULT_CONV_LAYER_CONFIG)
    encoder_embed_dim: int = 768  # model.encoder_embed_dim
    encoder_projection_dropout: float = 0  # model.dropout_input
    encoder_pos_conv_kernel: int = 95  # model.conv_pos
    encoder_pos_conv_groups: int = 16  # model.conv_pos_groups
    encoder_pos_conv_depth: int = 5  # model.pos_conv_depth
    encoder_num_layers: int = 12  # model.encoder_layers
    encoder_num_heads: int = 12  # model.encoder_attention_heads
    encoder_attention_dropout: float = 0.1  # model.attention_dropout
    encoder_ff_interm_features: int = 3072  # model.encoder_ffn_embed_dim
    encoder_ff_interm_dropout: float = 0.0  # model.activation_dropout
    encoder_dropout: float = 0.1  # model.dropout
    encoder_layer_norm_first: bool = False  # model.layer_norm_first
    encoder_layer_drop: float = 0.05  # model.encoder_layerdrop
    encoder_qkv_bias: bool = False  # WARNING: not in `fairseq`, always True there.

    codebook_size: int = 256  # model.codebook_size
    codebook_decay: float = 0.9  # model.codebook_decay
    num_codebooks: int = 8  # model.average_top_k_layers
    ema_start_decay: float = 0.999  # model.ema_decay
    ema_final_decay: float = 0.9999  # model.ema_end_decay
    ema_final_step: int = 30_000  # model.ema_anneal_end_step
    ema_exclude_layers: list[str] = field(default_factory=lambda: ["pos_conv_embed"])
    freeze_step: int = 200_000  # model.freeze_teacher_step


@dataclass(frozen=True)
class SpidRConfig(DinoSRConfig):
    """SpidR configuration. Corresponding names in `fairseq` in comments."""

    ema_timescale: float = 20_000
    ema_threshold: float = 1e-7
    encoder_layer_drop: float = 0.0  # model.encoder_layerdrop


@dataclass(frozen=True)
class MaskingConfig:
    """Masking configuration. Corresponding names in `fairseq` in comments."""

    mask_prob: float = 0.8  # model.mask_prob
    mask_selection: MaskingStrategy = "static"  # model.mask_selection
    mask_other: float = 0.0  # model.mask_other
    mask_length: int = 10  # model.mask_length
    no_mask_overlap: bool = False  # model.no_mask_overlap
    mask_min_space: int = 1  # model.mask_min_space
    mask_channel_prob: float = 0.0  # model.mask_channel_prob
    mask_channel_selection: MaskingStrategy = "static"  # model.mask_channel_selection
    mask_channel_other: float = 0.0  # model.mask_channel_other
    mask_channel_length: int = 10  # model.mask_channel_length
    no_mask_channel_overlap: bool = False  # model.no_mask_channel_overlap
    mask_channel_min_space: int = 1  # model.mask_channel_min_space


@dataclass(frozen=True)
class Config:
    """Full configuration."""

    run: RunConfig
    data: DataConfig
    model: DinoSRConfig | SpidRConfig
    optimizer: OptimizerConfig
    masking: MaskingConfig
    validation: dict[str, DataConfig]


@dataclass(frozen=True)
class ResumeConfig:
    """Hold information to resume from a checkpoint for validation."""

    step: int
    checkpoint: Path | None
    results: Path


def read_config(path: str | Path) -> Config:
    path = Path(path)
    if path.suffix not in {".json", ".toml"}:
        raise ValueError(f"Unsupported config file format: {path.suffix}. Use .json or .toml.")
    data = (tomllib.loads if path.suffix == ".toml" else json.loads)(Path(path).read_text(encoding="utf-8"))
    run = RunConfig(**data["run"])  # ty: ignore[missing-argument]
    return Config(
        run=run,
        data=DataConfig(**data["data"]),  # ty: ignore[missing-argument]
        model=(DinoSRConfig if run.model_type == "dinosr" else SpidRConfig)(**data.get("model", {})),
        optimizer=OptimizerConfig(**data.get("optimizer", {})),
        masking=MaskingConfig(**data.get("masking", {})),
        validation={k: DataConfig(**v) for k, v in data["validation"].items()} if "validation" in data else {},  # ty: ignore[missing-argument]
    )



# Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
"""Data loading and dataset utilities."""

from spidr.data.dataset import build_dataloader, speech_dataset
from spidr.data.utils import num_samples, read_manifest

__all__ = ["build_dataloader", "num_samples", "read_manifest", "speech_dataset"]
